from typing import Mapping


class LazyDict(Mapping):
    """Inspired by https://gist.github.com/gyli/9b50bb8537069b4e154fec41a4b5995a"""

    def __init__(self, *args, **kw):
        self._raw_dict = dict(*args, **kw)

    def __getitem__(self, key):
        func = self._raw_dict.__getitem__(key)
        return func()

    def __iter__(self):
        return iter(self._raw_dict)

    def __len__(self):
        return len(self._raw_dict)



from typing import Optional, Union

import numpy as np
import pandas as pd

from autora.variable import ValueType, VariableCollection


def pool(
    variables: VariableCollection,
    num_samples: int = 5,
    random_state: Optional[int] = None,
    replace: bool = True,
) -> pd.DataFrame:
    """
    Create a sequence of conditions randomly sampled from independent variables.

    Args:
        variables: the description of all the variables in the AER experiment.
        num_samples: the number of conditions to produce
        random_state: the seed value for the random number generator
        replace: if True, allow repeated values

    Returns: the generated conditions as a dataframe

    Examples:
        >>> from autora.state import State
        >>> from autora.variable import VariableCollection, Variable
        >>> from dataclasses import dataclass, field
        >>> import pandas as pd
        >>> import numpy as np

        With one independent variable "x", and some allowed_values we get some of those values
        back when running the experimentalist:
        >>> pool(
        ...     VariableCollection(
        ...         independent_variables=[Variable(name="x", allowed_values=range(10))
        ... ]), random_state=1)
           x
        0  4
        1  5
        2  7
        3  9
        4  0


        ... with one independent variable "x", and a value_range,
        we get a sample of the range back when running the experimentalist:
        >>> pool(
        ...     VariableCollection(independent_variables=[
        ...         Variable(name="x", value_range=(-5, 5))
        ... ]), random_state=1)
                  x
        0  0.118216
        1  4.504637
        2 -3.558404
        3  4.486494
        4 -1.881685



        The allowed_values or value_range must be specified:
        >>> pool(VariableCollection(independent_variables=[Variable(name="x")]))
        Traceback (most recent call last):
        ...
        ValueError: allowed_values or [value_range and type==REAL] needs to be set...

        With two independent variables, we get independent samples on both axes:
        >>> pool(VariableCollection(independent_variables=[
        ...         Variable(name="x1", allowed_values=range(1, 5)),
        ...         Variable(name="x2", allowed_values=range(1, 500)),
        ... ]), num_samples=10, replace=True, random_state=1)
           x1   x2
        0   2  434
        1   3  212
        2   4  137
        3   4  414
        4   1  129
        5   1  205
        6   4  322
        7   4  275
        8   1   43
        9   2   14

        If any of the variables have unspecified allowed_values, we get an error:
        >>> pool(
        ...     VariableCollection(independent_variables=[
        ...         Variable(name="x1", allowed_values=[1, 2]),
        ...         Variable(name="x2"),
        ... ]))
        Traceback (most recent call last):
        ...
        ValueError: allowed_values or [value_range and type==REAL] needs to be set...


        We can specify arrays of allowed values:

        >>> pool(
        ...     VariableCollection(independent_variables=[
        ...         Variable(name="x", allowed_values=np.linspace(-10, 10, 101)),
        ...         Variable(name="y", allowed_values=[3, 4]),
        ...         Variable(name="z", allowed_values=np.linspace(20, 30, 11)),
        ... ]), random_state=1)
             x  y     z
        0 -0.6  3  29.0
        1  0.2  4  24.0
        2  5.2  4  23.0
        3  9.0  3  29.0
        4 -9.4  3  22.0


    """
    rng = np.random.default_rng(random_state)

    raw_conditions = {}
    for iv in variables.independent_variables:
        if iv.allowed_values is not None:
            raw_conditions[iv.name] = rng.choice(
                iv.allowed_values, size=num_samples, replace=replace
            )
        elif (iv.value_range is not None) and (iv.type == ValueType.REAL):
            raw_conditions[iv.name] = rng.uniform(*iv.value_range, size=num_samples)

        else:
            raise ValueError(
                "allowed_values or [value_range and type==REAL] needs to be set for "
                "%s" % (iv)
            )

    return pd.DataFrame(raw_conditions)


random_pool = pool
"""Alias for `pool`"""


def sample(
    conditions: Union[pd.DataFrame, np.ndarray, np.recarray],
    num_samples: int = 1,
    random_state: Optional[int] = None,
    replace: bool = False,
) -> pd.DataFrame:
    """
    Take a random sample from some input conditions.

    Args:
        conditions: the conditions to sample from
        num_samples: the number of conditions to produce
        random_state: the seed value for the random number generator
        replace: if True, allow repeated values

    Returns: a Result object with a field `conditions` containing a DataFrame of the sampled
    conditions

    Examples:
        From a pd.DataFrame:
        >>> import pandas as pd
        >>> sample(
        ...     pd.DataFrame({"x": range(100, 200)}), num_samples=5, random_state=180)
              x
        67  167
        71  171
        64  164
        63  163
        96  196

        From a list (returns a DataFrame):
        >>> sample(range(1000), num_samples=5, random_state=180)
               0
        270  270
        908  908
        109  109
        331  331
        978  978
    """
    conditions_ = pd.DataFrame(conditions)
    return pd.DataFrame.sample(
        conditions_, random_state=random_state, n=num_samples, replace=replace
    )


random_sample = sample
"""Alias for `sample`"""



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional, Sequence, Tuple


class ValueType(str, Enum):
    """Specifies supported value types supported by Variables."""

    BOOLEAN = "boolean"
    INTEGER = "integer"
    REAL = "real"
    SIGMOID = "sigmoid"
    PROBABILITY = "probability"  # single probability
    PROBABILITY_SAMPLE = "probability_sample"  # sample from single probability
    PROBABILITY_DISTRIBUTION = (
        "probability_distribution"  # probability distribution over classes
    )
    CLASS = "class"  # sample from probability distribution over classes


@dataclass
class Variable:
    """Describes an experimental variable: name, type, range, units, and value of a variable."""

    name: str = ""
    value_range: Optional[Tuple[Any, Any]] = None
    allowed_values: Optional[Sequence] = None
    units: str = ""
    type: ValueType = ValueType.REAL
    variable_label: str = ""
    rescale: float = 1
    is_covariate: bool = False


@dataclass
class IV(Variable):
    """Independent variable."""

    name: str = "IV"
    variable_label: str = "Independent Variable"


@dataclass
class DV(Variable):
    """Dependent variable."""

    name: str = "DV"
    variable_label: str = "Dependent Variable"


@dataclass(frozen=True)
class VariableCollection:
    """Immutable metadata about dependent / independent variables and covariates."""

    independent_variables: Sequence[Variable] = field(default_factory=list)
    dependent_variables: Sequence[Variable] = field(default_factory=list)
    covariates: Sequence[Variable] = field(default_factory=list)


@dataclass
class IVTrial(IV):
    """
    Experiment trial as independent variable.
    """

    name: str = "trial"
    UID: str = ""
    variable_label: str = "Trial"
    units: str = "trials"
    priority: int = 0
    value_range: Tuple[Any, Any] = (0, 10000000)
    value: float = 0



from functools import partial
from itertools import product
from math import sqrt

import numpy as np

from autora.experimentalist.pipeline import (
    Pipeline,
    _parse_params_to_nested_dict,
    make_pipeline,
)

##############################################################################
# Building blocks
##############################################################################


def linear_pool_generator(stop=10):
    return range(stop)


def even_filter(values):
    return filter(lambda i: i % 2 == 0, values)


def odd_filter(values):
    return filter(lambda i: (i + 1) % 2 == 0, values)


def divisor_filter(values, divisor):
    return filter(lambda i: i % divisor == 0, values)


def is_sqrt_filter(values):
    return filter(lambda i: sqrt(i) % 1 == 0.0, values)


##############################################################################
# Simple pipelines of one variable
##############################################################################


def test_zeroth_pipline_zeroth_input():
    pipeline = Pipeline()
    result_0 = list(pipeline())
    assert result_0 == []


def test_zeroth_pipline_basic_input():
    pipeline = Pipeline([])
    result_0 = list(pipeline([0, 1, 2, 3]))
    assert result_0 == [0, 1, 2, 3]


def test_zeroth_make_pipeline():
    pipeline = make_pipeline()
    result = list(pipeline([0, 1, 2, 3]))
    assert result == [0, 1, 2, 3]


def test_single_element_pipeline():
    pipeline = Pipeline([("even_filter", even_filter)])
    result = list(pipeline(range(10)))
    assert result == [0, 2, 4, 6, 8]


def test_single_element_make_pipeline():
    pipeline = make_pipeline([even_filter])
    result = list(pipeline(range(10)))
    assert result == [0, 2, 4, 6, 8]


def test_multiple_element_pipeline():
    pipeline = Pipeline(
        [
            ("even_filter", even_filter),
            ("divisor_filter", partial(divisor_filter, divisor=3)),
        ]
    )
    result = list(pipeline(range(13)))
    assert result == [0, 6, 12]


def test_multiple_element_make_pipeline():
    pipeline = make_pipeline([even_filter, partial(divisor_filter, divisor=3)])
    result = list(pipeline(range(13)))
    assert result == [0, 6, 12]


def test_two_element_make_pipeline_with_params():
    pipeline = make_pipeline(
        [even_filter, divisor_filter], params={"divisor_filter": {"divisor": 5}}
    )
    result = list(pipeline(range(21)))
    assert result == [0, 10, 20]


def test_three_element_make_pipeline():
    pipeline = make_pipeline(
        [divisor_filter, divisor_filter, divisor_filter],
        params={
            "divisor_filter_0": {"divisor": 5},
            "divisor_filter_1": {"divisor": 7},
            "divisor_filter_2": {"divisor": 11},
        },
    )
    result = list(pipeline(range(500)))
    assert result == [0, 385]


def test_nested_pipeline():
    inner_pipeline = Pipeline([("pool", lambda: range(32))])
    outer_pipeline = Pipeline(
        [
            ("inner_pipeline", inner_pipeline),
            ("filter_by_divisor", partial(divisor_filter, divisor=8)),
        ],
    )
    result = list(outer_pipeline())
    assert result == [0, 8, 16, 24]


def test_nested_pipeline_nested_parameters():
    inner_pipeline = Pipeline([("pool", lambda maximum: range(maximum))])
    outer_pipeline = Pipeline(
        [
            ("inner_pipeline", inner_pipeline),
            ("filter_by_divisor", divisor_filter),
        ],
        params={
            "inner_pipeline": {"pool": {"maximum": 32}},
            "filter_by_divisor": {"divisor": 8},
        },
    )
    result = list(outer_pipeline())
    assert result == [0, 8, 16, 24]


def test_nested_pipeline_flat_parameters():
    inner_pipeline = Pipeline([("pool", lambda maximum: range(maximum))])
    outer_pipeline = Pipeline(
        [
            ("inner_pipeline", inner_pipeline),
            ("filter_by_divisor", divisor_filter),
        ],
        params={"inner_pipeline__pool__maximum": 32, "filter_by_divisor__divisor": 8},
    )
    result = list(outer_pipeline())
    assert result == [0, 8, 16, 24]


def test_nested_pipeline_nested_parameters_in_kwargs():
    inner_pipeline = Pipeline([("pool", lambda maximum: range(maximum))])
    outer_pipeline = Pipeline(
        [
            ("inner_pipeline", inner_pipeline),
            ("filter_by_divisor", divisor_filter),
        ],
    )
    result = list(
        outer_pipeline(
            **{
                "inner_pipeline": {"pool": {"maximum": 32}},
                "filter_by_divisor": {"divisor": 8},
            }
        )
    )
    assert result == [0, 8, 16, 24]


def test_nested_pipeline_flat_parameters_in_kwargs():
    inner_pipeline = Pipeline([("pool", lambda maximum: range(maximum))])
    outer_pipeline = Pipeline(
        [
            ("inner_pipeline", inner_pipeline),
            ("filter_by_divisor", divisor_filter),
        ],
    )
    result = list(
        outer_pipeline(
            **{"inner_pipeline__pool__maximum": 32, "filter_by_divisor__divisor": 8}
        )
    )
    assert result == [0, 8, 16, 24]


##############################################################################
# Simple pool and filters of one variable
##############################################################################


def test_zeroth_poolpipeline():
    pipeline = make_pipeline([linear_pool_generator])
    result = list(pipeline())
    assert result == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_even_poolpipeline():
    pipeline = make_pipeline([linear_pool_generator, even_filter])
    result = list(pipeline())
    assert result == [0, 2, 4, 6, 8]


def test_odd_poolpipeline():
    pipeline = make_pipeline([linear_pool_generator, odd_filter])
    result = list(pipeline())
    assert result == [1, 3, 5, 7, 9]


def test_poolpipeline_run():
    pipeline = make_pipeline([linear_pool_generator, odd_filter])
    result = list(pipeline.run())
    assert result == [1, 3, 5, 7, 9]


##############################################################################
# Pool and filters of two Weber variables
##############################################################################


def weber_pool(vmin=0, vmax=1, steps=5):
    s1 = s2 = np.linspace(vmin, vmax, steps)
    pool = product(s1, s2)
    return pool


def weber_filter(values):
    return filter(lambda s: s[0] >= s[1], values)


def test_weber_unfiltered_poolpipeline():
    pipeline = make_pipeline([weber_pool])
    result = list(pipeline())
    assert result[0] == (0.0, 0.0)
    assert result[1] == (0.0, 0.25)
    assert result[-1] == (1.0, 1.0)


def test_weber_filtered_poolpipeline():
    pipeline = make_pipeline([weber_pool, weber_filter])
    result = list(pipeline())
    assert result[0] == (0.0, 0.0)
    assert result[1] == (0.25, 0.0)
    assert result[-1] == (1.0, 1.0)


##############################################################################
# Helper Functions
##############################################################################
def test_params_parser_zero_level():
    params = {
        "model": "%%newest_theory%%",
        "n": 10,
        "measure": "least_confident",
    }
    result = _parse_params_to_nested_dict(params, divider="__")
    assert result == params


def test_params_parser_one_level():
    params = {
        "pool__ivs": "%%independent_variables%%",
        "uncertainty_sampler__model": "%%newest_theory%%",
        "uncertainty_sampler__n": 10,
        "uncertainty_sampler__measure": "least_confident",
    }

    result = _parse_params_to_nested_dict(params, divider="__")
    assert result == {
        "pool": {
            "ivs": "%%independent_variables%%",
        },
        "uncertainty_sampler": {
            "model": "%%newest_theory%%",
            "n": 10,
            "measure": "least_confident",
        },
    }


def test_params_parser_recurse_one():
    params = {
        "filter_pipeline__step1__n_samples": 100,
    }

    result = _parse_params_to_nested_dict(params, divider="__")
    assert result == {"filter_pipeline": {"step1": {"n_samples": 100}}}


def test_params_parser_recurse_one_n_levels():
    params = {
        "a__b__c__d__e__f": 100,
        "a__b__c__d__e__g": 200,
        "a__b__h": 300,
    }
    result = _parse_params_to_nested_dict(params, divider="__")
    assert result == {"a": {"b": {"c": {"d": {"e": {"f": 100, "g": 200}}}, "h": 300}}}


def test_params_parser_recurse_one_n_levels_alternative_divider():
    params = {
        "a:b:c:d:e:f": 100,
        "a:b:c:d:e:g": 200,
        "a:b:h": 300,
    }
    result = _parse_params_to_nested_dict(params, divider=":")
    assert result == {"a": {"b": {"c": {"d": {"e": {"f": 100, "g": 200}}}, "h": 300}}}


def test_params_parser_recurse():
    params = {
        "pool__ivs": "%%independent_variables%%",
        "filter_pipeline__step1__n_samples": 100,
        "filter_pipeline__step2__n_samples": 10,
        "uncertainty_sampler__model": "%%newest_theory%%",
        "uncertainty_sampler__n": 10,
        "uncertainty_sampler__measure": "least_confident",
    }

    result = _parse_params_to_nested_dict(params, divider="__")
    assert result == {
        "pool": {
            "ivs": "%%independent_variables%%",
        },
        "filter_pipeline": {"step1": {"n_samples": 100}, "step2": {"n_samples": 10}},
        "uncertainty_sampler": {
            "model": "%%newest_theory%%",
            "n": 10,
            "measure": "least_confident",
        },
    }


##############################################################################
# Parallel Pipelines
##############################################################################


def test_parallelpipeline_run():
    pl = make_pipeline([range(3), range(10, 13)], kind="union")
    assert list(pl.run()) == [0, 1, 2, 10, 11, 12]


def test_parallelpipeline_many_steps():
    pl = make_pipeline([range(0, 5) for _ in range(1000)], kind="union")
    results = list(pl.run())
    assert len(results) == 5000
    assert results[0:10] == [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]



import yaml


def dump(data, file):
    yaml.dump(data, file, Dumper=yaml.Dumper)
    return


def load(file):
    data = yaml.load(file, Loader=yaml.Loader)
    return data


def dumps(data):
    string = yaml.dump(data, Dumper=yaml.Dumper)
    return string


def loads(string):
    data = yaml.load(string, Loader=yaml.Loader)
    return data



import logging
import pathlib
import tempfile
from typing import Optional

import numpy as np
from hypothesis import Verbosity, given, settings
from hypothesis import strategies as st

from autora.serializer import SupportedSerializer, load_state
from autora.state import StandardState
from autora.workflow.__main__ import main

_logger = logging.getLogger(__name__)

example_workflow_library_module = st.sampled_from(["_example_workflow_library"])


def validate_model(state: Optional[StandardState]):
    assert state is not None

    assert state.conditions is not None
    assert len(state.conditions) == 100

    assert state.experiment_data is not None
    assert len(state.experiment_data) == 100

    assert state.models[-1] is not None
    assert np.allclose(state.models[-1].coef_, [[2.0]])
    assert np.allclose(state.models[-1].intercept_, [[0.5]])


@given(example_workflow_library_module)
def test_e2e_nominal(workflow_library_module):
    """Test a basic standard chain of CLI calls using the default serializer.

    Equivalent to:
    $ python -m autora.workflow test_workflow.initial_state --out-path start
    $ python -m autora.workflow test_workflow.experimentalist --in-path start --out-path conditions
    $ python -m autora.workflow test_workflow.experiment_runner --in-path conditions --out-path data
    $ python -m autora.workflow test_workflow.theorist --in-path data --out-path theory
    """

    with tempfile.TemporaryDirectory() as d:
        main(
            f"{workflow_library_module}.initial_state",
            out_path=pathlib.Path(d, "start"),
        )
        main(
            f"{workflow_library_module}.experimentalist",
            in_path=pathlib.Path(d, "start"),
            out_path=pathlib.Path(d, "conditions"),
        )
        main(
            f"{workflow_library_module}.experiment_runner",
            in_path=pathlib.Path(d, "conditions"),
            out_path=pathlib.Path(d, "data"),
        )
        main(
            f"{workflow_library_module}.theorist",
            in_path=pathlib.Path(d, "data"),
            out_path=pathlib.Path(d, "theory"),
        )

        final_state = load_state(pathlib.Path(d, "theory"))
        validate_model(final_state)


@given(
    example_workflow_library_module,
    st.sampled_from(SupportedSerializer),
    st.booleans(),
    st.booleans(),
)
@settings(verbosity=Verbosity.verbose, deadline=1000)
def test_e2e_serializers(workflow_library_module, serializer, verbose, debug):
    """Test a basic standard chain of CLI calls using a single serializer."""

    common_settings = dict(
        in_loader=serializer, out_dumper=serializer, verbose=verbose, debug=debug
    )

    with tempfile.TemporaryDirectory() as d:
        main(
            f"{workflow_library_module}.initial_state",
            out_path=pathlib.Path(d, "start"),
            **common_settings,
        )
        main(
            f"{workflow_library_module}.experimentalist",
            in_path=pathlib.Path(d, "start"),
            out_path=pathlib.Path(d, "conditions"),
            **common_settings,
        )
        main(
            f"{workflow_library_module}.experiment_runner",
            in_path=pathlib.Path(d, "conditions"),
            out_path=pathlib.Path(d, "data"),
            **common_settings,
        )
        main(
            f"{workflow_library_module}.theorist",
            in_path=pathlib.Path(d, "data"),
            out_path=pathlib.Path(d, "theory"),
            **common_settings,
        )

        final_state: StandardState = load_state(
            pathlib.Path(d, "theory"), loader=serializer
        )
        validate_model(final_state)


@given(
    example_workflow_library_module,
    st.sampled_from(SupportedSerializer),
    st.sampled_from(SupportedSerializer),
    st.sampled_from(SupportedSerializer),
    st.sampled_from(SupportedSerializer),
    st.booleans(),
    st.booleans(),
)
@settings(verbosity=Verbosity.verbose, deadline=1000)
def test_e2e_valid_serializer_mix(
    workflow_library_module,
    initial_serializer,
    experimental_serializer,
    experiment_runner_serializer,
    theorist_serializer,
    verbose,
    debug,
):
    """Test a basic standard chain of CLI calls using a mix of serializers."""

    common_settings = dict(verbose=verbose, debug=debug)

    with tempfile.TemporaryDirectory() as d:
        main(
            f"{workflow_library_module}.initial_state",
            out_path=pathlib.Path(d, "start"),
            out_dumper=initial_serializer,
            **common_settings,
        )
        main(
            f"{workflow_library_module}.experimentalist",
            in_path=pathlib.Path(d, "start"),
            out_path=pathlib.Path(d, "conditions"),
            in_loader=initial_serializer,
            out_dumper=experimental_serializer,
            **common_settings,
        )
        main(
            f"{workflow_library_module}.experiment_runner",
            in_path=pathlib.Path(d, "conditions"),
            out_path=pathlib.Path(d, "data"),
            in_loader=experimental_serializer,
            out_dumper=experiment_runner_serializer,
            **common_settings,
        )
        main(
            f"{workflow_library_module}.theorist",
            in_path=pathlib.Path(d, "data"),
            out_path=pathlib.Path(d, "theory"),
            in_loader=experiment_runner_serializer,
            out_dumper=theorist_serializer,
            **common_settings,
        )

        final_state: StandardState = load_state(
            pathlib.Path(d, "theory"), loader=theorist_serializer
        )
        validate_model(final_state)



import time

from autora.variable import DV, IV


class VTime:
    """
    A class representing time as a general experimental variable.
    """

    _t0 = 0

    def __init__(self):
        """
        Initializes the time.
        """
        self._t0 = time.time()

    # Resets reference time.
    def reset(self):
        """
        Resets the time.
        """
        self._t0 = time.time()


class IVTime(IV, VTime):
    """
    A class representing time as an independent variable.
    """

    _name = "time_IV"
    _UID = ""
    _variable_label = "Time"
    _units = "s"
    _priority = 0
    _value_range = (0, 3600)
    _value = 0

    # Initializes reference time.
    # The reference time usually denotes the beginning of an experiment trial.
    def __init__(self, *args, **kwargs):
        """
        Initializes the time as independent variable.

        For arguments, see [autora.variable.Variable][autora.variable.Variable.__init__]
        """
        super(IVTime, self).__init__(*args, **kwargs)

    # Waits until specified time has passed relative to reference time
    def manipulate(self):
        """
        Waits for the specified time to pass.
        """

        t_wait = self.get_value() - (time.time() - self._t0)
        if t_wait <= 0:
            return
        else:
            time.sleep(t_wait)

    def disconnect(self):
        """
        Disconnects the time.
        """
        pass


class DVTime(DV, VTime):
    """
    A class representing time as a dependent variable.
    """

    _name = "time_DV"
    _UID = ""
    _variable_label = "Time"
    _units = "s"
    _priority = 0
    _value_range = (0, 604800)  # don't record more than a week
    _value = 0

    _is_covariate = True

    # Initializes reference time.
    # The reference time usually denotes the beginning of an experiment trial.
    def __init__(self, *args, **kwargs):
        """
        Initializes the time as dependent variable. The reference time usually denotes
        the beginning of an experiment trial.

        For arguments, see [autora.variable.Variable][autora.variable.Variable.__init__]
        """
        print(self._variable_label)
        super(DVTime, self).__init__(*args, **kwargs)
        print(self._variable_label)

    # Measure number of seconds relative to reference time
    def measure(self):
        """
        Measures the time in seconds relative to the reference time.
        """
        value = time.time() - self._t0
        self.set_value(value)



import logging

import pandas as pd
from hypothesis import HealthCheck, given, settings

from autora.state import StandardState

from .test_serializer import serializer_dump_load_strategy
from .test_strategies import standard_state_strategy

logger = logging.getLogger(__name__)


@given(standard_state_strategy(), serializer_dump_load_strategy)
@settings(suppress_health_check={HealthCheck.too_slow}, deadline=1000)
def test_state_serialize_deserialize(o: StandardState, dump_load):
    o_loaded = dump_load(o)
    assert o.variables == o_loaded.variables
    assert pd.DataFrame.equals(o.conditions, o_loaded.conditions)
    assert pd.DataFrame.equals(o.experiment_data, o_loaded.experiment_data)



"""Classes to represent cycle state $S$ as $S_n = S_{0} + \\sum_{i=1}^n \\Delta S_{i}$."""

from __future__ import annotations

import copy
import inspect
import logging
import warnings
from collections import UserDict
from dataclasses import dataclass, field, fields, is_dataclass, replace
from enum import Enum
from functools import singledispatch, wraps
from typing import (
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator

from autora.variable import VariableCollection

_logger = logging.getLogger(__name__)


T = TypeVar("T")
C = TypeVar("C", covariant=True)


class DeltaAddable(Protocol[C]):
    """A class which a Delta or other Mapping can be added to, returning the same class"""

    def __add__(self: C, other: Union[Delta, Mapping]) -> C:
        ...


S = TypeVar("S", bound=DeltaAddable)


@dataclass(frozen=True)
class State:
    """
    Base object for dataclasses which use the Delta mechanism.

    Examples:
        >>> from dataclasses import dataclass, field
        >>> from typing import List, Optional

        We define a dataclass where each field (which is going to be delta-ed) has additional
        metadata "delta" which describes its delta behaviour.
        >>> @dataclass(frozen=True)
        ... class ListState(State):
        ...    l: List = field(default_factory=list, metadata={"delta": "extend"})
        ...    m: List = field(default_factory=list, metadata={"delta": "replace"})

        Now we instantiate the dataclass...
        >>> l = ListState(l=list("abc"), m=list("xyz"))
        >>> l
        ListState(l=['a', 'b', 'c'], m=['x', 'y', 'z'])

        ... and can add deltas to it. `l` will be extended:
        >>> l + Delta(l=list("def"))
        ListState(l=['a', 'b', 'c', 'd', 'e', 'f'], m=['x', 'y', 'z'])

        ... wheras `m` will be replaced:
        >>> l + Delta(m=list("uvw"))
        ListState(l=['a', 'b', 'c'], m=['u', 'v', 'w'])

        ... they can be chained:
        >>> l + Delta(l=list("def")) + Delta(m=list("uvw"))
        ListState(l=['a', 'b', 'c', 'd', 'e', 'f'], m=['u', 'v', 'w'])

        ... and we update multiple fields with one Delta:
        >>> l + Delta(l=list("ghi"), m=list("rst"))
        ListState(l=['a', 'b', 'c', 'g', 'h', 'i'], m=['r', 's', 't'])

        A non-existent field will be ignored:
        >>> l + Delta(o="not a field")
        ListState(l=['a', 'b', 'c'], m=['x', 'y', 'z'])

        ... but will trigger a warning:
        >>> with warnings.catch_warnings(record=True) as w:
        ...     _ = l + Delta(o="not a field")
        ...     print(w[0].message) # doctest: +NORMALIZE_WHITESPACE
        These fields: ['o'] could not be used to update ListState,
        which has these fields & aliases: ['l', 'm']

        We can also use the `.update` method to do the same thing:
        >>> l.update(l=list("ghi"), m=list("rst"))
        ListState(l=['a', 'b', 'c', 'g', 'h', 'i'], m=['r', 's', 't'])

        We can also define fields which `append` the last result:
        >>> @dataclass(frozen=True)
        ... class AppendState(State):
        ...    n: List = field(default_factory=list, metadata={"delta": "append"})

        >>> m = AppendState(n=list("ɑβɣ"))
        >>> m
        AppendState(n=['ɑ', 'β', 'ɣ'])

        `n` will be appended:
        >>> m + Delta(n="∂")
        AppendState(n=['ɑ', 'β', 'ɣ', '∂'])

        The metadata key "converter" is used to coerce types (inspired by
        [PEP 712](https://peps.python.org/pep-0712/)):
        >>> @dataclass(frozen=True)
        ... class CoerceStateList(State):
        ...    o: Optional[List] = field(default=None, metadata={"delta": "replace"})
        ...    p: List = field(default_factory=list, metadata={"delta": "replace",
        ...                                                    "converter": list})

        >>> r = CoerceStateList()

        If there is no `metadata["converter"]` set for a field, no coercion occurs
        >>> r + Delta(o="not a list")
        CoerceStateList(o='not a list', p=[])

        If there is a `metadata["converter"]` set for a field, the data are coerced:
        >>> r + Delta(p="not a list")
        CoerceStateList(o=None, p=['n', 'o', 't', ' ', 'a', ' ', 'l', 'i', 's', 't'])

        If the input data are of the correct type, they are returned unaltered:
        >>> r + Delta(p=["a", "list"])
        CoerceStateList(o=None, p=['a', 'list'])

        With a converter, inputs are converted to the type output by the converter:
        >>> @dataclass(frozen=True)
        ... class CoerceStateDataFrame(State):
        ...    q: pd.DataFrame = field(default_factory=pd.DataFrame,
        ...                            metadata={"delta": "replace",
        ...                                      "converter": pd.DataFrame})

        If the type is already correct, the object is passed to the converter,
        but should be returned unchanged:
        >>> s = CoerceStateDataFrame()
        >>> (s + Delta(q=pd.DataFrame([("a",1,"alpha"), ("b",2,"beta")], columns=list("xyz")))).q
           x  y      z
        0  a  1  alpha
        1  b  2   beta

        If the type is not correct, the object is converted if possible. For a dataframe,
        we can convert records:
        >>> (s + Delta(q=[("a",1,"alpha"), ("b",2,"beta")])).q
           0  1      2
        0  a  1  alpha
        1  b  2   beta

        ... or an array:
        >>> (s + Delta(q=np.linspace([1, 2], [10, 15], 3))).q
              0     1
        0   1.0   2.0
        1   5.5   8.5
        2  10.0  15.0

        ... or a dictionary:
        >>> (s + Delta(q={"a": [1,2,3], "b": [4,5,6]})).q
           a  b
        0  1  4
        1  2  5
        2  3  6

        ... or a list:
        >>> (s + Delta(q=[11, 12, 13])).q
            0
        0  11
        1  12
        2  13

        ... but not, for instance, a string:
        >>> (s + Delta(q="not compatible with pd.DataFrame")).q
        Traceback (most recent call last):
        ...
        ValueError: DataFrame constructor not properly called!

        Without a converter:
        >>> @dataclass(frozen=True)
        ... class CoerceStateDataFrameNoConverter(State):
        ...    r: pd.DataFrame = field(default_factory=pd.DataFrame, metadata={"delta": "replace"})

        ... there is no coercion – the object is passed unchanged
        >>> t = CoerceStateDataFrameNoConverter()
        >>> (t + Delta(r=np.linspace([1, 2], [10, 15], 3))).r
        array([[ 1. ,  2. ],
               [ 5.5,  8.5],
               [10. , 15. ]])


        A converter can cast from a DataFrame to a np.ndarray (with a single datatype),
        for instance:
        >>> @dataclass(frozen=True)
        ... class CoerceStateArray(State):
        ...    r: Optional[np.ndarray] = field(default=None,
        ...                            metadata={"delta": "replace",
        ...                                      "converter": np.asarray})

        Here we pass a dataframe, but expect a numpy array:
        >>> (CoerceStateArray() + Delta(r=pd.DataFrame([("a",1), ("b",2)], columns=list("xy")))).r
        array([['a', 1],
               ['b', 2]], dtype=object)

        We can define aliases which can transform between different potential field
        names.

    """

    def __add__(self, other: Union[Delta, Mapping]):
        updates = dict()
        other_fields_unused = list(other.keys())
        for self_field in fields(self):
            other_value, key = _get_value(self_field, other)
            if other_value is None:
                continue
            other_fields_unused.remove(key)

            self_field_key = self_field.name
            self_value = getattr(self, self_field_key)
            delta_behavior = self_field.metadata["delta"]

            if (constructor := self_field.metadata.get("converter", None)) is not None:
                coerced_other_value = constructor(other_value)
            else:
                coerced_other_value = other_value

            if delta_behavior == "extend":
                extended_value = _extend(self_value, coerced_other_value)
                updates[self_field_key] = extended_value
            elif delta_behavior == "append":
                appended_value = _append(self_value, coerced_other_value)
                updates[self_field_key] = appended_value
            elif delta_behavior == "replace":
                updates[self_field_key] = coerced_other_value
            else:
                raise NotImplementedError(
                    "delta_behaviour=`%s` not implemented" % delta_behavior
                )

        if len(other_fields_unused) > 0:
            warnings.warn(
                "These fields: %s could not be used to update %s, "
                "which has these fields & aliases: %s"
                % (
                    other_fields_unused,
                    type(self).__name__,
                    _get_field_names_and_aliases(self),
                ),
            )

        new = replace(self, **updates)
        return new

    def update(self, **kwargs):
        """
        Return a new version of the State with values updated.

        This is identical to adding a `Delta`.

        If you need to replace values, ignoring the State value aggregation rules,
        use `dataclasses.replace` instead.
        """
        return self + Delta(**kwargs)


def _get_value(f, other: Union[Delta, Mapping]):
    """
    Given a `State`'s `dataclasses.field` f, get a value from `other` and report its name.

    Returns: a tuple (the value, the key associated with that value)

    Examples:
        >>> from dataclasses import field, dataclass, fields
        >>> @dataclass
        ... class Example:
        ...     a: int = field()

        For a field with no aliases, we retrieve values with the base name:
        >>> f_a = fields(Example)[0]
        >>> _get_value(f_a, Delta(a=1))
        (1, 'a')

        ... and only the base name:
        >>> print(_get_value(f_a, Delta(b=2)))  # no match for b
        (None, None)

        Any other names are unimportant:
        >>> _get_value(f_a, Delta(b=2, a=1))
        (1, 'a')

    """

    key = f.name

    value, used_key = None, None

    if key in other.keys():
        value = other[key]
        used_key = key

    return value, used_key


def _get_field_names_and_aliases(s: State):
    """
    Get a list of field names and their aliases from a State object

    Args:
        s: a State object

    Returns: a list of field names and their aliases on `s`

    Examples:
        >>> from dataclasses import field
        >>> @dataclass(frozen=True)
        ... class SomeState(State):
        ...    l: List = field(default_factory=list)
        ...    m: List = field(default_factory=list)
        >>> _get_field_names_and_aliases(SomeState())
        ['l', 'm']

    """
    result = [f.name for f in fields(s)]
    return result


class Delta(UserDict, Generic[S]):
    """
    Represents a delta where the base object determines the extension behavior.

    Examples:
        >>> from dataclasses import dataclass

        First we define the dataclass to act as the basis:
        >>> from typing import Optional, List
        >>> @dataclass(frozen=True)
        ... class ListState:
        ...     l: Optional[List] = None
        ...     m: Optional[List] = None
        ...
    """

    pass


Result = Delta
"""`Result` is an alias for `Delta`."""


@singledispatch
def _extend(a, b):
    """
    Function to extend supported datatypes.

    """
    raise NotImplementedError("`_extend` not implemented for %s, %s" % (a, b))


@_extend.register(type(None))
def _extend_none(_, b):
    """
    Implementation of `_extend` to support None-types.

    Examples:
        >>> _extend(None, [])
        []

        >>> _extend(None, [3])
        [3]
    """
    return b


@_extend.register(list)
def _extend_list(a, b):
    """
    Implementation of `_extend` to support Lists.

    Examples:
        >>> _extend([], [])
        []

        >>> _extend([1,2], [3])
        [1, 2, 3]
    """
    return a + b


@_extend.register(pd.DataFrame)
def _extend_pd_dataframe(a, b):
    """
    Implementation of `_extend` to support DataFrames.

    Examples:
        >>> _extend(pd.DataFrame({"a": []}), pd.DataFrame({"a": []}))
        Empty DataFrame
        Columns: [a]
        Index: []

        >>> _extend(pd.DataFrame({"a": [1,2,3]}), pd.DataFrame({"a": [4,5,6]}))
           a
        0  1
        1  2
        2  3
        3  4
        4  5
        5  6
    """
    return pd.concat((a, b), ignore_index=True)


@_extend.register(np.ndarray)
def _extend_np_ndarray(a, b):
    """
    Implementation of `_extend` to support Numpy ndarrays.

    Examples:
        >>> _extend(np.array([(1,2,3), (4,5,6)]), np.array([(7,8,9)]))
        array([[1, 2, 3],
               [4, 5, 6],
               [7, 8, 9]])
    """
    return np.row_stack([a, b])


@_extend.register(dict)
def _extend_dict(a, b):
    """
    Implementation of `_extend` to support Dictionaries.

    Examples:
        >>> _extend({"a": "cats"}, {"b": "dogs"})
        {'a': 'cats', 'b': 'dogs'}
    """
    return dict(a, **b)


@_extend.register(Delta)
def _extend_delta(a, b):
    """
    Implementation of `_extend` to support Delta

    Examples:
        >>> _extend(Delta(conditions=[1,2]), Delta(conditions=[3,4]))
        {'conditions': [1, 2, 3, 4]}
    """
    res = copy.deepcopy(a)
    for key in b:
        if key in a:
            res[key] = _extend(a[key], b[key])
        else:
            res[key] = b[key]
    return Delta(res)


def _append(a: List[T], b: T) -> List[T]:
    """
    Function to create a new list with an item appended to it.

    Examples:
        Given a starting list `a_`:
        >>> a_ = [1, 2, 3]

        ... we can append a value:
        >>> _append(a_, 4)
        [1, 2, 3, 4]

        `a_` is unchanged
        >>> a_ == [1, 2, 3]
        True

        Why not just use `list.append`? `list.append` mutates `a` in place, which we can't allow
        in the AER cycle – parts of the cycle rely on purely functional code which doesn't
        (accidentally or intentionally) manipulate existing data.
        >>> list.append(a_, 4)  # not what we want
        >>> a_
        [1, 2, 3, 4]
    """
    return a + [b]


def inputs_from_state(f, input_mapping: Dict = {}):
    """Decorator to make target `f` into a function on a `State` and `**kwargs`.

    This wrapper makes it easier to pass arguments to a function from a State.

    It was inspired by the pytest "fixtures" mechanism.

    Args:
        f: a function with arguments that could be fields on a `State`
            and that returns a `Delta`.
        input_mapping: a dict that maps the input arguments of the function to the state fields

    Returns: a version of `f` which takes and returns `State` objects.

    Examples:
        >>> from dataclasses import dataclass, field
        >>> import pandas as pd
        >>> from typing import List, Optional

        The `State` it operates on needs to have the metadata described in the state module:
        >>> @dataclass(frozen=True)
        ... class U(State):
        ...     conditions: List[int] = field(metadata={"delta": "replace"})

        We indicate the inputs required by the parameter names.
        The output must be (compatible with) a `Delta` object.
        >>> @inputs_from_state
        ... def experimentalist(conditions):
        ...     new_conditions = [c + 10 for c in conditions]
        ...     return new_conditions

        >>> experimentalist(U(conditions=[1,2,3,4]))
        [11, 12, 13, 14]

        >>> experimentalist(U(conditions=[101,102,103,104]))
        [111, 112, 113, 114]

        If our function uses a different keyword argument than the state field, we can use
        the input mapping:
        >>> def experimentalist_(X):
        ...     new_conditions = [x + 10 for x in X]
        ...     return new_conditions
        >>> experimentalist_on_state = inputs_from_state(experimentalist_, {'X': 'conditions'})
        >>> experimentalist_on_state(U(conditions=[1,2,3,4]))
        [11, 12, 13, 14]

        Both also work with the `State` as UserDict. Here, we use the StandardState
        >>> experimentalist(StandardState(conditions=[1, 2, 3, 4]))
        [11, 12, 13, 14]

        >>> experimentalist_on_state(StandardState(conditions=[1, 2, 3, 4]))
        [11, 12, 13, 14]

        A dictionary can be returned and used:
        >>> @inputs_from_state
        ... def returns_a_dictionary(conditions):
        ...     new_conditions = [c + 10 for c in conditions]
        ...     return {"conditions": new_conditions}
        >>> returns_a_dictionary(U(conditions=[2]))
        {'conditions': [12]}

        >>> from autora.variable import VariableCollection, Variable
        >>> from sklearn.base import BaseEstimator
        >>> from sklearn.linear_model import LinearRegression

        >>> @inputs_from_state
        ... def theorist(experiment_data: pd.DataFrame, variables: VariableCollection, **kwargs):
        ...     ivs = [vi.name for vi in variables.independent_variables]
        ...     dvs = [vi.name for vi in variables.dependent_variables]
        ...     X, y = experiment_data[ivs], experiment_data[dvs]
        ...     model = LinearRegression(fit_intercept=True).set_params(**kwargs).fit(X, y)
        ...     return model

        >>> @dataclass(frozen=True)
        ... class V(State):
        ...     variables: VariableCollection  # field(metadata={"delta":... }) omitted ∴ immutable
        ...     experiment_data: pd.DataFrame = field(metadata={"delta": "extend"})
        ...     model: Optional[BaseEstimator] = field(metadata={"delta": "replace"}, default=None)

        >>> v = V(
        ...     variables=VariableCollection(independent_variables=[Variable("x")],
        ...                                  dependent_variables=[Variable("y")]),
        ...     experiment_data=pd.DataFrame({"x": [0,1,2,3,4], "y": [2,3,4,5,6]})
        ... )
        >>> model = theorist(v)
        >>> model.coef_, model.intercept_
        (array([[1.]]), array([2.]))

        Arguments from the state can be overridden by passing them in as keyword arguments (kwargs):
        >>> theorist(v, experiment_data=pd.DataFrame({"x": [0,1,2,3], "y": [12,13,14,15]}))\\
        ...     .intercept_
        array([12.])

        ... and other arguments supported by the inner function can also be passed
        (if and only if the inner function allows for and handles `**kwargs` arguments alongside
        the values from the state).
        >>> theorist(v, fit_intercept=False).intercept_
        0.0

        Any parameters not provided by the state must be provided by default values or by the
        caller. If the default is specified:
        >>> @inputs_from_state
        ... def experimentalist(conditions, offset=25):
        ...     new_conditions = [c + offset for c in conditions]
        ...     return new_conditions

        ... then it need not be passed.
        >>> experimentalist(U(conditions=[1,2,3,4]))
        [26, 27, 28, 29]

        If a default isn't specified:
        >>> @inputs_from_state
        ... def experimentalist(conditions, offset):
        ...     new_conditions = [c + offset for c in conditions]
        ...     return new_conditions

        ... then calling the experimentalist without it will throw an error:
        >>> experimentalist(U(conditions=[1,2,3,4]))
        Traceback (most recent call last):
        ...
        TypeError: experimentalist() missing 1 required positional argument: 'offset'

        ... which can be fixed by passing the argument as a keyword to the wrapped function.
        >>> experimentalist(U(conditions=[1,2,3,4]), offset=2)
        [3, 4, 5, 6]

        The same is true, if we don't provide a mapping for arguments:
        >>> def experimentalist_(X, offset):
        ...     new_conditions = [x + offset for x in X]
        ...     return new_conditions
        >>> experimentalist_on_state = inputs_from_state(experimentalist_, {'X': 'conditions'})
        >>> experimentalist_on_state(StandardState(conditions=[1,2,3,4]), offset=2)
        [3, 4, 5, 6]

        The state itself is passed through if the inner function requests the `state`:
        >>> @inputs_from_state
        ... def function_which_needs_whole_state(state, conditions):
        ...     print("Doing something on: ", state)
        ...     new_conditions = [c + 2 for c in conditions]
        ...     return new_conditions
        >>> function_which_needs_whole_state(U(conditions=[1,2,3,4]))
        Doing something on:  U(conditions=[1, 2, 3, 4])
        [3, 4, 5, 6]

    """
    # Get the set of parameter names from function f's signature

    reversed_mapping = {v: k for k, v in input_mapping.items()}

    parameters_ = set(inspect.signature(f).parameters.keys())
    missing_func_params = set(input_mapping.keys()).difference(parameters_)
    if missing_func_params:
        raise ValueError(
            f"The following keys in input_state_mapping are not parameters of the function: "
            f"{missing_func_params}"
        )

    @wraps(f)
    def _f(state_: S, /, **kwargs) -> S:
        # Get the parameters needed which are available from the state_.
        # All others must be provided as kwargs or default values on f.
        assert is_dataclass(state_) or isinstance(state_, UserDict)
        if is_dataclass(state_):
            from_state = parameters_.intersection({i.name for i in fields(state_)})
            arguments_from_state = {k: getattr(state_, k) for k in from_state}
            from_state_input_mapping = {
                reversed_mapping.get(field.name, field.name): getattr(
                    state_, field.name
                )
                for field in fields(state_)
                if reversed_mapping.get(field.name, field.name) in parameters_
            }
            arguments_from_state.update(from_state_input_mapping)
        elif isinstance(state_, UserDict):
            from_state = parameters_.intersection(set(state_.keys()))
            arguments_from_state = {k: state_[k] for k in from_state}
            from_state_input_mapping = {
                reversed_mapping.get(key, key): state_[key]
                for key in state_.keys()
                if reversed_mapping.get(key, key) in parameters_
            }
            arguments_from_state.update(from_state_input_mapping)
        if "state" in parameters_:
            arguments_from_state["state"] = state_
        arguments = dict(arguments_from_state, **kwargs)
        result = f(**arguments)
        return result

    return _f


def outputs_to_delta(*output: str):
    """
    Decorator factory to wrap outputs from a function as Deltas.

    Examples:
        >>> @outputs_to_delta("conditions")
        ... def add_five(x):
        ...     return [xi + 5 for xi in x]

        >>> add_five([1, 2, 3])
        {'conditions': [6, 7, 8]}

        >>> @outputs_to_delta("c")
        ... def add_six(conditions):
        ...     return [c + 5 for c in conditions]

        >>> add_six([1, 2, 3])
        {'c': [6, 7, 8]}

        >>> @outputs_to_delta("+1", "-1")
        ... def plus_minus_1(x):
        ...     a = [xi + 1 for xi in x]
        ...     b = [xi - 1 for xi in x]
        ...     return a, b

        >>> plus_minus_1([1, 2, 3])
        {'+1': [2, 3, 4], '-1': [0, 1, 2]}


        If the wrong number of values are specified for the return, then there might be errors.
        If multiple outputs are expected, but only a single output is returned, we get a warning:
        >>> @outputs_to_delta("1", "2")
        ... def returns_single_result_when_more_expected():
        ...     return "a"
        >>> returns_single_result_when_more_expected()  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Traceback (most recent call last):
        ...
        AssertionError: function `<function returns_single_result_when_more_expected at 0x...>`
        has to return multiple values to match `('1', '2')`. Got `a` instead.

        If multiple outputs are expected, but the wrong number are returned, we get a warning:
        >>> @outputs_to_delta("1", "2", "3")
        ... def returns_wrong_number_of_results():
        ...     return "a", "b"
        >>> returns_wrong_number_of_results()  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Traceback (most recent call last):
        ...
        AssertionError: function `<function returns_wrong_number_of_results at 0x...>`
        has to return exactly `3` values to match `('1', '2', '3')`. Got `('a', 'b')` instead.

        However, if a single output is expected, and multiple are returned, these are treated as
        a single object and no error occurs:
        >>> @outputs_to_delta("foo")
        ... def returns_a_tuple():
        ...     return "a", "b", "c"
        >>> returns_a_tuple()
        {'foo': ('a', 'b', 'c')}

        If we fail to specify output names, an error is returned immediately.
        >>> @outputs_to_delta()
        ... def decorator_missing_arguments():
        ...     return "a", "b", "c"
        Traceback (most recent call last):
        ...
        ValueError: `output` names must be specified.

    """

    def decorator(f):
        if len(output) == 0:
            raise ValueError("`output` names must be specified.")

        elif len(output) == 1:

            @wraps(f)
            def inner(*args, **kwargs):
                result = f(*args, **kwargs)
                delta = Delta(**{output[0]: result})
                return delta

        else:

            @wraps(f)
            def inner(*args, **kwargs):
                result = f(*args, **kwargs)
                assert isinstance(result, tuple), (
                    "function `%s` has to return multiple values "
                    "to match `%s`. Got `%s` instead." % (f, output, result)
                )
                assert len(output) == len(result), (
                    "function `%s` has to return "
                    "exactly `%s` values "
                    "to match `%s`. "
                    "Got `%s` instead."
                    "" % (f, len(output), output, result)
                )
                delta = Delta(**dict(zip(output, result)))
                return delta

        return inner

    return decorator


def delta_to_state(f):
    """Decorator to make `f` which takes a `State` and returns a `Delta` return an updated `State`.

    This wrapper handles adding a returned Delta to an input State object.

    Args:
        f: the function which returns a `Delta` object

    Returns: the function modified to return a State object

    Examples:
        >>> from dataclasses import dataclass, field
        >>> import pandas as pd
        >>> from typing import List, Optional

        The `State` it operates on needs to have the metadata described in the state module:
        >>> @dataclass(frozen=True)
        ... class U(State):
        ...     conditions: List[int] = field(metadata={"delta": "replace"})

        We indicate the inputs required by the parameter names.
        The output must be (compatible with) a `Delta` object.
        >>> @delta_to_state
        ... @inputs_from_state
        ... def experimentalist(conditions):
        ...     new_conditions = [c + 10 for c in conditions]
        ...     return Delta(conditions=new_conditions)

        >>> experimentalist(U(conditions=[1,2,3,4]))
        U(conditions=[11, 12, 13, 14])

        >>> experimentalist(U(conditions=[101,102,103,104]))
        U(conditions=[111, 112, 113, 114])

        If the output of the function is not a `Delta` object (or something compatible with its
        interface), then an error is thrown.
        >>> @delta_to_state
        ... @inputs_from_state
        ... def returns_bare_conditions(conditions):
        ...     new_conditions = [c + 10 for c in conditions]
        ...     return new_conditions

        >>> returns_bare_conditions(U(conditions=[1])) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        AssertionError: Output of <function returns_bare_conditions at 0x...> must be a `Delta`,
        `UserDict`, or `dict`.

        A dictionary can be returned and used:
        >>> @delta_to_state
        ... @inputs_from_state
        ... def returns_a_dictionary(conditions):
        ...     new_conditions = [c + 10 for c in conditions]
        ...     return {"conditions": new_conditions}
        >>> returns_a_dictionary(U(conditions=[2]))
        U(conditions=[12])

        ... as can an object which subclasses UserDict (like `Delta`)
        >>> class MyDelta(UserDict):
        ...     pass
        >>> @delta_to_state
        ... @inputs_from_state
        ... def returns_a_userdict(conditions):
        ...     new_conditions = [c + 10 for c in conditions]
        ...     return MyDelta(conditions=new_conditions)
        >>> returns_a_userdict(U(conditions=[3]))
        U(conditions=[13])

        We recommend using the `Delta` object rather than a `UserDict` or `dict` as its
        functionality may be expanded in future.

        >>> from autora.variable import VariableCollection, Variable
        >>> from sklearn.base import BaseEstimator
        >>> from sklearn.linear_model import LinearRegression

        >>> @delta_to_state
        ... @inputs_from_state
        ... def theorist(experiment_data: pd.DataFrame, variables: VariableCollection, **kwargs):
        ...     ivs = [vi.name for vi in variables.independent_variables]
        ...     dvs = [vi.name for vi in variables.dependent_variables]
        ...     X, y = experiment_data[ivs], experiment_data[dvs]
        ...     new_model = LinearRegression(fit_intercept=True).set_params(**kwargs).fit(X, y)
        ...     return Delta(model=new_model)

        >>> @dataclass(frozen=True)
        ... class V(State):
        ...     variables: VariableCollection  # field(metadata={"delta":... }) omitted ∴ immutable
        ...     experiment_data: pd.DataFrame = field(metadata={"delta": "extend"})
        ...     model: Optional[BaseEstimator] = field(metadata={"delta": "replace"}, default=None)

        >>> v = V(
        ...     variables=VariableCollection(independent_variables=[Variable("x")],
        ...                                  dependent_variables=[Variable("y")]),
        ...     experiment_data=pd.DataFrame({"x": [0,1,2,3,4], "y": [2,3,4,5,6]})
        ... )
        >>> v_prime = theorist(v)
        >>> v_prime.model.coef_, v_prime.model.intercept_
        (array([[1.]]), array([2.]))

        Arguments from the state can be overridden by passing them in as keyword arguments (kwargs):
        >>> theorist(v, experiment_data=pd.DataFrame({"x": [0,1,2,3], "y": [12,13,14,15]}))\\
        ...     .model.intercept_
        array([12.])

        ... and other arguments supported by the inner function can also be passed
        (if and only if the inner function allows for and handles `**kwargs` arguments alongside
        the values from the state).
        >>> theorist(v, fit_intercept=False).model.intercept_
        0.0

        Any parameters not provided by the state must be provided by default values or by the
        caller. If the default is specified:
        >>> @delta_to_state
        ... @inputs_from_state
        ... def experimentalist(conditions, offset=25):
        ...     new_conditions = [c + offset for c in conditions]
        ...     return Delta(conditions=new_conditions)

        ... then it need not be passed.
        >>> experimentalist(U(conditions=[1,2,3,4]))
        U(conditions=[26, 27, 28, 29])

        If a default isn't specified:
        >>> @delta_to_state
        ... @inputs_from_state
        ... def experimentalist(conditions, offset):
        ...     new_conditions = [c + offset for c in conditions]
        ...     return Delta(conditions=new_conditions)

        ... then calling the experimentalist without it will throw an error:
        >>> experimentalist(U(conditions=[1,2,3,4]))
        Traceback (most recent call last):
        ...
        TypeError: experimentalist() missing 1 required positional argument: 'offset'

        ... which can be fixed by passing the argument as a keyword to the wrapped function.
        >>> experimentalist(U(conditions=[1,2,3,4]), offset=2)
        U(conditions=[3, 4, 5, 6])

        The state itself is passed through if the inner function requests the `state`:
        >>> @delta_to_state
        ... @inputs_from_state
        ... def function_which_needs_whole_state(state, conditions):
        ...     print("Doing something on: ", state)
        ...     new_conditions = [c + 2 for c in conditions]
        ...     return Delta(conditions=new_conditions)
        >>> function_which_needs_whole_state(U(conditions=[1,2,3,4]))
        Doing something on:  U(conditions=[1, 2, 3, 4])
        U(conditions=[3, 4, 5, 6])

    """

    @wraps(f)
    def _f(state_: S, **kwargs) -> S:
        delta = f(state_, **kwargs)
        assert isinstance(delta, Mapping), (
            "Output of %s must be a `Delta`, `UserDict`, " "or `dict`." % f
        )
        new_state = state_ + delta
        return new_state

    return _f


def on_state(
    function: Optional[Callable] = None,
    input_mapping: Dict = {},
    output: Optional[Sequence[str]] = None,
):
    """Decorator (factory) to make target `function` into a function on a `State` and `**kwargs`.

    This combines the functionality of `outputs_to_delta` and `inputs_from_state`

    Args:
        function: the function to be wrapped
        output: list specifying State field names for the return values of `function`
        input_mapping: a dict that maps the keywords of the functions to the state fields

    Returns:

    Examples:
        >>> from dataclasses import dataclass, field
        >>> import pandas as pd
        >>> from typing import List, Optional

        The `State` it operates on needs to have the metadata described in the state module:
        >>> @dataclass(frozen=True)
        ... class W(State):
        ...     conditions: List[int] = field(metadata={"delta": "replace"})

        We indicate the inputs required by the parameter names.
        >>> def add_ten(conditions):
        ...     return [c + 10 for c in conditions]
        >>> experimentalist = on_state(function=add_ten, output=["conditions"])

        >>> experimentalist(W(conditions=[1,2,3,4]))
        W(conditions=[11, 12, 13, 14])

        You can wrap functions which return a Delta object natively, by omitting the `output`
        argument:
        >>> @on_state()
        ... def add_five(conditions):
        ...     return Delta(conditions=[c + 5 for c in conditions])

        >>> add_five(W(conditions=[1, 2, 3, 4]))
        W(conditions=[6, 7, 8, 9])

        If you fail to declare outputs for a function which doesn't return a Delta:
        >>> @on_state()
        ... def missing_output_param(conditions):
        ...     return [c + 5 for c in conditions]

        ... an exception is raised:
        >>> missing_output_param(W(conditions=[1])) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        AssertionError: Output of <function missing_output_param at 0x...> must be a `Delta`,
        `UserDict`, or `dict`.

        You can use the @on_state(output=[...]) as a decorator:
        >>> @on_state(output=["conditions"])
        ... def add_six(conditions):
        ...     return [c + 6 for c in conditions]

        >>> add_six(W(conditions=[1, 2, 3, 4]))
        W(conditions=[7, 8, 9, 10])

        You can also declare an input-to-output mapping if the keyword arguments of the functions
        don't match the state fields:
        >>> @on_state(input_mapping={'X': 'conditions'}, output=["conditions"])
        ... def add_six(X):
        ...     return [x + 6 for x in X]

        >>> add_six(W(conditions=[1, 2, 3, 4]))
        W(conditions=[7, 8, 9, 10])

        This also works on the StandardState or other States that are defined as UserDicts:
        >>> add_six(StandardState(conditions=[1, 2, 3,4])).conditions
            0
        0   7
        1   8
        2   9
        3  10
    """

    def decorator(f):
        f_ = f
        if output is not None:
            f_ = outputs_to_delta(*output)(f_)
        f_ = inputs_from_state(f_, input_mapping)
        f_ = delta_to_state(f_)
        return f_

    if function is None:
        return decorator
    else:
        return decorator(function)


StateFunction = Callable[[State], State]


class StandardStateVariables(Enum):
    CONDITIONS = "conditions"
    EXPERIMENT_DATA = "experiment_data"
    MODELS = "models"
    VARIABLES = "variables"


@dataclass(frozen=True)
class StandardState(State):
    """
    Examples:
        The state can be initialized emtpy
        >>> from autora.variable import VariableCollection, Variable
        >>> s = StandardState()
        >>> s
        StandardState(variables=None, conditions=None, experiment_data=None, models=[])

        The `variables` can be updated using a `Delta`:
        >>> dv1 = Delta(variables=VariableCollection(independent_variables=[Variable("1")]))
        >>> s + dv1 # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        StandardState(variables=VariableCollection(independent_variables=[Variable(name='1',...)

        ... and are replaced by each `Delta`:
        >>> dv2 = Delta(variables=VariableCollection(independent_variables=[Variable("2")]))
        >>> s + dv1 + dv2 # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        StandardState(variables=VariableCollection(independent_variables=[Variable(name='2',...)

        The `conditions` can be updated using a `Delta`:
        >>> dc1 = Delta(conditions=pd.DataFrame({"x": [1, 2, 3]}))
        >>> (s + dc1).conditions
           x
        0  1
        1  2
        2  3

        ... and are replaced by each `Delta`:
        >>> dc2 = Delta(conditions=pd.DataFrame({"x": [4, 5]}))
        >>> (s + dc1 + dc2).conditions
           x
        0  4
        1  5

        Datatypes other than `pd.DataFrame` will be coerced into a `DataFrame` if possible.
        >>> import numpy as np
        >>> dc3 = Delta(conditions=np.core.records.fromrecords([(8, "h"), (9, "i")], names="n,c"))
        >>> (s + dc3).conditions
           n  c
        0  8  h
        1  9  i

        If they are passed without column names, no column names are inferred.
        This is to ensure that accidental mislabeling of columns cannot occur.
        Column names should usually be provided.
        >>> dc4 = Delta(conditions=[(6,), (7,)])
        >>> (s + dc4).conditions
           0
        0  6
        1  7

        Datatypes which are incompatible with a pd.DataFrame will throw an error:
        >>> s + Delta(conditions="not compatible with pd.DataFrame") \
# doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: ...

        Experiment data can be updated using a Delta:
        >>> ded1 = Delta(experiment_data=pd.DataFrame({"x": [1,2,3], "y": ["a", "b", "c"]}))
        >>> (s + ded1).experiment_data
           x  y
        0  1  a
        1  2  b
        2  3  c

        ... and are extended with each Delta:
        >>> ded2 = Delta(experiment_data=pd.DataFrame({"x": [4, 5, 6], "y": ["d", "e", "f"]}))
        >>> (s + ded1 + ded2).experiment_data
           x  y
        0  1  a
        1  2  b
        2  3  c
        3  4  d
        4  5  e
        5  6  f

        If they are passed without column names, no column names are inferred.
        This is to ensure that accidental mislabeling of columns cannot occur.
        >>> ded3 = Delta(experiment_data=pd.DataFrame([(7, "g"), (8, "h")]))
        >>> (s + ded3).experiment_data
           0  1
        0  7  g
        1  8  h

        If there are already data present, the column names must match.
        >>> (s + ded2 + ded3).experiment_data
             x    y    0    1
        0  4.0    d  NaN  NaN
        1  5.0    e  NaN  NaN
        2  6.0    f  NaN  NaN
        3  NaN  NaN  7.0    g
        4  NaN  NaN  8.0    h

        `experiment_data` other than `pd.DataFrame` will be coerced into a `DataFrame` if possible.
        >>> import numpy as np
        >>> ded4 = Delta(
        ...     experiment_data=np.core.records.fromrecords([(1, "a"), (2, "b")], names=["x", "y"]))
        >>> (s + ded4).experiment_data
           x  y
        0  1  a
        1  2  b

        `experiment_data` which are incompatible with a pd.DataFrame will throw an error:
        >>> s + Delta(experiment_data="not compatible with pd.DataFrame") \
# doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: ...

        `models` can be updated using a Delta:
        >>> from sklearn.dummy import DummyClassifier
        >>> dm1 = Delta(models=[DummyClassifier(constant=1)])
        >>> dm2 = Delta(models=[DummyClassifier(constant=2), DummyClassifier(constant=3)])
        >>> (s + dm1).models
        [DummyClassifier(constant=1)]

        >>> (s + dm1 + dm2).models
        [DummyClassifier(constant=1), DummyClassifier(constant=2), DummyClassifier(constant=3)]

    """

    variables: Optional[VariableCollection] = field(
        default=None, metadata={"delta": "replace"}
    )
    conditions: Optional[pd.DataFrame] = field(
        default=None, metadata={"delta": "replace", "converter": pd.DataFrame}
    )
    experiment_data: Optional[pd.DataFrame] = field(
        default=None, metadata={"delta": "extend", "converter": pd.DataFrame}
    )
    models: List[BaseEstimator] = field(
        default_factory=list,
        metadata={"delta": "extend"},
    )


X = TypeVar("X")
Y = TypeVar("Y")
XY = TypeVar("XY")


def estimator_on_state(estimator: BaseEstimator) -> StateFunction:
    """
    Convert a scikit-learn compatible estimator into a function on a `State` object.

    Supports passing additional `**kwargs` which are used to update the estimator's params
    before fitting.

    Examples:
        Initialize a function which operates on the state, `state_fn` and runs a LinearRegression.
        >>> from sklearn.linear_model import LinearRegression
        >>> state_fn = estimator_on_state(LinearRegression())

        Define the state on which to operate (here an instance of the `StandardState`):
        >>> from autora.state import StandardState
        >>> from autora.variable import Variable, VariableCollection
        >>> import pandas as pd
        >>> s = StandardState(
        ...     variables=VariableCollection(
        ...         independent_variables=[Variable("x")],
        ...         dependent_variables=[Variable("y")]),
        ...     experiment_data=pd.DataFrame({"x": [1,2,3], "y":[3,6,9]})
        ... )

        Run the function, which fits the model and adds the result to the `StandardState` as the
        last entry in the .models list.
        >>> state_fn(s).models[-1].coef_
        array([[3.]])

    """

    @on_state()
    def theorist(
        experiment_data: pd.DataFrame, variables: VariableCollection, **kwargs
    ):
        ivs = [v.name for v in variables.independent_variables]
        dvs = [v.name for v in variables.dependent_variables]
        X, y = experiment_data[ivs], experiment_data[dvs]
        new_model = estimator.set_params(**kwargs).fit(X, y)
        return Delta(models=[new_model])

    return theorist


def experiment_runner_on_state(f: Callable[[X], XY]) -> StateFunction:
    """Wrapper for experiment_runner of the form $f(x) \rarrow (x,y)$, where `f`
    returns both $x$ and $y$ values in a complete dataframe.

    Examples:
        The conditions are some x-values in a StandardState object:
        >>> from autora.state import StandardState
        >>> s = StandardState(conditions=pd.DataFrame({"x": [1, 2, 3]}))

        The function can be defined on a DataFrame, allowing the explicit inclusion of
        metadata like column names.
        >>> def x_to_xy_fn(c: pd.DataFrame) -> pd.Series:
        ...     result = c.assign(y=lambda df: 2 * df.x + 1)
        ...     return result

        We apply the wrapped function to `s` and look at the returned experiment_data:
        >>> experiment_runner_on_state(x_to_xy_fn)(s).experiment_data
           x  y
        0  1  3
        1  2  5
        2  3  7

        We can also define functions of several variables:
        >>> def xs_to_xy_fn(c: pd.DataFrame) -> pd.Series:
        ...     result = c.assign(y=c.x0 + c.x1)
        ...     return result

        With the relevant variables as conditions:
        >>> t = StandardState(conditions=pd.DataFrame({"x0": [1, 2, 3], "x1": [10, 20, 30]}))
        >>> experiment_runner_on_state(xs_to_xy_fn)(t).experiment_data
           x0  x1   y
        0   1  10  11
        1   2  20  22
        2   3  30  33

    """

    @on_state()
    def experiment_runner(conditions: pd.DataFrame, **kwargs):
        x = conditions
        experiment_data = f(x, **kwargs)
        return Delta(experiment_data=experiment_data)

    return experiment_runner


def combined_functions_on_state(
    functions: List[Tuple[str, Callable]], output: Optional[Sequence[str]] = None
):
    """
    Decorator (factory) to make target list of `functions` into a function on a `State`.
    The resulting function uses a state field as input and combines the outputs of the
    `functions`.

    Args:
        functions: the list of functions to be wrapped
        output: list specifying State field names for the return values of `function`

    Examples:
        >>> @dataclass(frozen=True)
        ... class U(State):
        ...     conditions: List[int] = field(metadata={"delta": "replace"})
        >>> identity = lambda conditions : conditions
        >>> double_conditions = combined_functions_on_state(
        ...     [('id_1', identity), ('id_2', identity)], output=['conditions'])
        >>> s = U([1, 2])
        >>> s_double = double_conditions(s)
        >>> s
        U(conditions=[1, 2])
        >>> s_double
        U(conditions=[1, 2, 1, 2])

        # We can also pass parameters to the functions:
        >>> def multiply(conditions, multiplier):
        ...     return [el * multiplier for el in conditions]
        >>> double_and_triple = combined_functions_on_state(
        ...     [('doubler', multiply), ('tripler', multiply)], output=['conditions']
        ... )
        >>> s = U([1, 2])
        >>> s_double_triple = double_and_triple(
        ...     s, params={'doubler': {'multiplier': 2}, 'tripler': {'multiplier': 3}}
        ... )
        >>> s_double_triple
        U(conditions=[2, 4, 3, 6])

        # If the functions return a Delta object, we don't need to provide an output argument
        >>> def decrement(conditions, dec):
        ...     return Delta(conditions=[el-dec for el in conditions])
        >>> def increment(conditions, inc):
        ...     return Delta(conditions=[el+inc for el in conditions])
        >>> dec_and_inc = combined_functions_on_state(
        ...     [('decrement', decrement), ('increment', increment)])
        >>> s_dec_and_inc = dec_and_inc(
        ...     s, params={'decrement': {'dec': 10}, 'increment': {'inc': 2}})
        >>> s_dec_and_inc
        U(conditions=[-9, -8, 3, 4])

    """

    def f_(_state: State, params: Optional[Dict] = None):
        result_delta = None
        for name, function in functions:
            _f_input_from_state = inputs_from_state(function)
            if params is None:
                _params = {}
            else:
                _params = params
            if name in _params.keys():
                _delta = _f_input_from_state(_state, **_params[name])
            else:
                _delta = _f_input_from_state(_state)
            result_delta = _extend(result_delta, _delta)
        return result_delta

    if output:
        f_ = outputs_to_delta(*output)(f_)
    f_ = delta_to_state(f_)
    return f_



import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures

from autora.experimentalist.grid import grid_pool
from autora.state import StandardState, estimator_on_state, on_state
from autora.variable import Variable, VariableCollection

rng = np.random.default_rng()


def initial_state(_):
    state = StandardState(
        variables=VariableCollection(
            independent_variables=[
                Variable(name="x", allowed_values=np.linspace(-10, +10, 1001))
            ],
            dependent_variables=[Variable(name="y")],
            covariates=[],
        ),
        conditions=None,
        experiment_data=pd.DataFrame({"x": [], "y": []}),
        models=[],
    )
    return state


@on_state(output=["conditions"])
def experimentalist(variables):
    conditions: pd.DataFrame = grid_pool(variables)
    selected_conditions = conditions.sample(10, random_state=rng)
    return selected_conditions


coefs = [2.0, 3.0, 1.0]
noise_std = 10.0


def ground_truth(x, coefs_=coefs):
    return coefs_[0] * x**2.0 + coefs_[1] * x + coefs_[2]


@on_state(output=["experiment_data"])
def experiment_runner(conditions, coefs_=coefs, noise_std_=noise_std, rng=rng):
    experiment_data = conditions.assign(
        y=(
            ground_truth(conditions["x"], coefs_=coefs_)
            + rng.normal(0.0, noise_std_, size=conditions["x"].shape)
        )
    )
    return experiment_data


theorist = estimator_on_state(
    GridSearchCV(
        make_pipeline(PolynomialFeatures(), LinearRegression()),
        param_grid={"polynomialfeatures__degree": [0, 1, 2, 3, 4]},
        scoring="r2",
    )
)



from __future__ import annotations

import collections

import numpy as np
import numpy.typing


def sequence_to_array(iterable):
    """
    Converts a finite sequence of experimental conditions into a 2D numpy.array.

    See also: [array_to_sequence][autora.experimentalist.utils.array_to_sequence]

    Examples:

        A simple range object can be converted into an array of dimension 2:
        >>> sequence_to_array(range(5)) # doctest: +NORMALIZE_WHITESPACE
        array([[0], [1], [2], [3], [4]])

        For mixed datatypes, the highest-level type common to all the inputs will be used, so
        consider using [_sequence_to_recarray][autora.experimentalist.utils._sequence_to_recarray]
        instead.
        >>> sequence_to_array(zip(range(5), "abcde"))  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        array([['0', 'a'], ['1', 'b'], ['2', 'c'], ['3', 'd'], ['4', 'e']],  dtype='<U...')

        Single strings are broken into characters:
        >>> sequence_to_array("abcde")  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        array([['a'], ['b'], ['c'], ['d'], ['e']], dtype='<U...')

        Multiple strings are treated as individual entries:
        >>> sequence_to_array(["abc", "de"])  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        array([['abc'], ['de']], dtype='<U...')

    """
    deque = collections.deque(iterable)
    array = np.array(deque).reshape((len(deque), -1))
    return array


def sequence_to_recarray(iterable):
    """
    Converts a finite sequence of experimental conditions into a numpy recarray.

    See also: [array_to_sequence][autora.experimentalist.utils.array_to_sequence]

    Examples:

        A simple range object is converted into a recarray of dimension 2:
        >>> sequence_to_recarray(range(5)) # doctest: +NORMALIZE_WHITESPACE
        rec.array([(0,), (1,), (2,), (3,), (4,)], dtype=[('f0', '<i...')])

        Mixed datatypes lead to multiple output types:
        >>> sequence_to_recarray(zip(range(5), "abcde"))  # doctest: +NORMALIZE_WHITESPACE
        rec.array([(0, 'a'), (1, 'b'), (2, 'c'), (3, 'd'), (4, 'e')],
            dtype=[('f0', '<i...'), ('f1', '<U1')])

        Single strings are broken into characters:
        >>> sequence_to_recarray("abcde")  # doctest: +NORMALIZE_WHITESPACE
        rec.array([('a',), ('b',), ('c',), ('d',), ('e',)], dtype=[('f0', '<U1')])

        Multiple strings are treated as individual entries:
        >>> sequence_to_recarray(["abc", "de"])  # doctest: +NORMALIZE_WHITESPACE
        rec.array([('abc',), ('de',)], dtype=[('f0', '<U3')])

    """
    deque = collections.deque(iterable)

    if isinstance(deque[0], (str, int, float, complex)):
        recarray = np.core.records.fromrecords([(d,) for d in deque])
    else:
        recarray = np.core.records.fromrecords(deque)

    return recarray


def array_to_sequence(input: numpy.typing.ArrayLike):
    """
    Convert an array of experimental conditions into an iterable of smaller arrays.

    See also:
        - [sequence_to_array][autora.experimentalist.utils.sequence_to_array]
        - [sequence_to_array][autora.experimentalist.utils.sequence_to_recarray]

    Examples:

        We start with an array:
        >>> a0 = np.arange(10).reshape(-1,2)
        >>> a0
        array([[0, 1],
               [2, 3],
               [4, 5],
               [6, 7],
               [8, 9]])

        The sequence is created as a generator object
        >>> array_to_sequence(a0)  # doctest: +ELLIPSIS
        <generator object array_to_sequence at 0x...>

        To see the sequence, we can convert it into a list:
        >>> l0 = list(array_to_sequence(a0))
        >>> l0
        [array([0, 1]), array([2, 3]), array([4, 5]), array([6, 7]), array([8, 9])]

        The individual rows are themselves 1-dimensional arrays:
        >>> l0[0]
        array([0, 1])

        The rows can be subscripted as usual:
        >>> l0[2][1]
        5

        We can also use a record array:
        >>> a1 = np.rec.fromarrays([range(5), list("abcde")])
        >>> a1
        rec.array([(0, 'a'), (1, 'b'), (2, 'c'), (3, 'd'), (4, 'e')],
                  dtype=[('f0', '<i...'), ('f1', '<U1')])

        This is converted into records:
        >>> l1 = list(array_to_sequence(a1))
        >>> l1
        [(0, 'a'), (1, 'b'), (2, 'c'), (3, 'd'), (4, 'e')]

        The elements of the list are numpy.records
        >>> type(l1[0])
        <class 'numpy.record'>

    """
    assert isinstance(input, (np.ndarray, np.recarray))

    for a in input:
        yield a



"""
Provides tools to chain functions used to create experiment sequences.
"""
from __future__ import annotations

import copy
from itertools import chain
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    get_args,
    runtime_checkable,
)


@runtime_checkable
class Pool(Protocol):
    """Creates an experimental sequence from scratch."""

    def __call__(self) -> _ExperimentalSequence:
        ...


@runtime_checkable
class Pipe(Protocol):
    """Takes in an _ExperimentalSequence and modifies it before returning it."""

    def __call__(self, ex: _ExperimentalSequence) -> _ExperimentalSequence:
        ...


_StepType = Tuple[str, Union[Pool, Pipe, Iterable]]
_StepType.__doc__ = (
    "A Pipeline step's name and generating object, as tuple(name, pipeline_piece)."
)

PARAM_DIVIDER = "__"


class Pipeline:
    """
    Processes ("pipelines") a series of ExperimentalSequences through a pipeline.

    Examples:
        A pipeline which filters even values 0 to 9:
        >>> p = Pipeline(
        ... [("is_even", lambda values: filter(lambda i: i % 2 == 0, values))]  # a "pipe" function
        ... )
        >>> list(p(range(10)))
        [0, 2, 4, 6, 8]

        A pipeline which filters for square, odd numbers:
        >>> from math import sqrt
        >>> p = Pipeline([
        ... ("is_odd", lambda values: filter(lambda i: i % 2 != 0, values)),
        ... ("is_sqrt", lambda values: filter(lambda i: sqrt(i) % 1 == 0., values))
        ... ])
        >>> list(p(range(100)))
        [1, 9, 25, 49, 81]


        >>> from itertools import product
        >>> Pipeline([("pool", lambda: product(range(5), ["a", "b"]))]) # doctest: +ELLIPSIS
        Pipeline(steps=[('pool', <function <lambda> at 0x...>)], params={})

        >>> Pipeline([
        ... ("pool", lambda: product(range(5), ["a", "b"])),
        ... ("filter", lambda values: filter(lambda i: i[0] % 2 == 0, values))
        ... ]) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Pipeline(steps=[('pool', <function <lambda> at 0x...>), \
        ('filter', <function <lambda> at 0x...>)], \
        params={})

        >>> pipeline = Pipeline([
        ... ("pool", lambda maximum: product(range(maximum), ["a", "b"])),
        ... ("filter", lambda values, divisor: filter(lambda i: i[0] % divisor == 0, values))
        ... ] ,
        ... params = {"pool": {"maximum":5}, "filter": {"divisor": 2}})
        >>> pipeline # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Pipeline(steps=[('pool', <function <lambda> at 0x...>), \
        ('filter', <function <lambda> at 0x...>)], \
        params={'pool': {'maximum': 5}, 'filter': {'divisor': 2}})
        >>> list(pipeline.run())
        [(0, 'a'), (0, 'b'), (2, 'a'), (2, 'b'), (4, 'a'), (4, 'b')]

        >>> pipeline.params = {"pool": {"maximum":7}, "filter": {"divisor": 3}}
        >>> list(pipeline())
        [(0, 'a'), (0, 'b'), (3, 'a'), (3, 'b'), (6, 'a'), (6, 'b')]

        >>> pipeline.params = {"pool": {"maximum":7}}
        >>> list(pipeline()) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: <lambda>() missing 1 required positional argument: 'divisor'


    """

    def __init__(
        self,
        steps: Optional[Sequence[_StepType]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the pipeline with a series of Pipe objects."""
        if steps is None:
            steps = list()
        self.steps = steps

        if params is None:
            params = dict()
        self.params = params

    def __repr__(self):
        return f"{self.__class__.__name__}(steps={self.steps}, params={self.params})"

    def __call__(
        self,
        ex: Optional[_ExperimentalSequence] = None,
        **params,
    ) -> _ExperimentalSequence:
        """Successively pass the input values through the Pipe."""

        # Initialize the parameters objects.
        merged_params = self._merge_params_with_self_params(params)

        try:
            # Check we have steps to use
            assert len(self.steps) > 0
        except AssertionError:
            # If the pipeline doesn't have any steps...
            if ex is not None:
                # ...the output is the input
                return ex
            elif ex is None:
                # ... unless the input was None, in which case it's an emtpy list
                return []

        # Make an iterator from the steps, so that we can be sure to only go through them once
        # (Otherwise if we handle the "pool" as a special case, we have to track our starting point)
        pipes_iterator = iter(self.steps)

        # Initialize our results object
        if ex is None:
            # ... there's no input, so presumably the first element in the steps is a pool
            # which should generate our initial values.
            name, pool = next(pipes_iterator)
            if isinstance(pool, Pool):
                # Here, the pool is a Pool callable, which we can pass parameters.
                all_params_for_pool = merged_params.get(name, dict())
                results = [pool(**all_params_for_pool)]
            elif isinstance(pool, Iterable):
                # Otherwise, the pool should be an iterable which we can just use as is.
                results = [pool]

        else:
            # ... there's some input, so we can use that as the initial value
            results = [ex]

        # Run the successive steps over the last result
        for name, pipe in pipes_iterator:
            assert isinstance(pipe, Pipe)
            all_params_for_pipe = merged_params.get(name, dict())
            results.append(pipe(results[-1], **all_params_for_pipe))

        return results[-1]

    def _merge_params_with_self_params(self, params):
        pipeline_params = _parse_params_to_nested_dict(
            self.params, divider=PARAM_DIVIDER
        )
        call_params = _parse_params_to_nested_dict(params, divider=PARAM_DIVIDER)
        merged_params = _merge_dicts(pipeline_params, call_params)
        return merged_params

    run = __call__


def _merge_dicts(a: dict, b: dict):
    """
    merges b into a.

    Args:
        a: the "base" dictionary
        b: the "update" dictionary which takes precendence

    Returns: the combined dictionary

    Originally from https://stackoverflow.com/a/7205107, modified for AER to allow overwriting.

    Examples:
        Non-conflicting dictionaries are merged "side-by-side"
        >>> _merge_dicts({1:{"a":"A"},2:{"b":"B"}}, {2:{"c":"C"},3:{"d":"D"}})
        {1: {'a': 'A'}, 2: {'b': 'B', 'c': 'C'}, 3: {'d': 'D'}}

        With conflicting dictionaries, the second dictionary takes precedence
        >>> _merge_dicts(
        ...     {"l1_a": {"l2_1": {"l3_alpha": "from_first"}}},
        ...     {"l1_a": {"l2_1": {"l3_alpha": "from_second"}}})
        {'l1_a': {'l2_1': {'l3_alpha': 'from_second'}}}

        Again, with non-conflicting dictionaries at the lower level
        >>> _merge_dicts(
        ...     {"l1_a": {"l2_1": {"l3_alpha": "from_first"}}},
        ...     {"l1_a": {"l2_1": {"l3_beta": "from_second"}}})
        {'l1_a': {'l2_1': {'l3_alpha': 'from_first', 'l3_beta': 'from_second'}}}

        >>> _merge_dicts(
        ...     {"l1_a": {"l2_1": {"l3_alpha": "from_first", "l3_beta": "from_first"}}},
        ...     {"l1_a": {"l2_1": {                          "l3_beta": "from_second"}}})
        {'l1_a': {'l2_1': {'l3_alpha': 'from_first', 'l3_beta': 'from_second'}}}

    """
    a_, b_ = dict(a), dict(b)

    for key in b_:
        if key in a_:
            if isinstance(a_[key], dict) and isinstance(b_[key], dict):
                a_[key] = _merge_dicts(a_[key], b_[key])
            elif a_[key] != b_[key]:
                a_[key] = b_[key]
            else:
                pass
        else:
            a_[key] = b_[key]
    return a_


class PipelineUnion(Pipeline):
    """
    Run several Pipes in parallel and concatenate all their results.

    Examples:
        You can use the ParallelPipeline to parallelize a group of poolers:
        >>> union_pipeline_0 = PipelineUnion([
        ...      ("pool_1", make_pipeline([range(5)])),
        ...      ("pool_2", make_pipeline([range(25, 30)])),
        ...     ]
        ... )
        >>> list(union_pipeline_0.run())
        [0, 1, 2, 3, 4, 25, 26, 27, 28, 29]

        >>> union_pipeline_1 = PipelineUnion([
        ...      ("pool_1", range(5)),
        ...      ("pool_2", range(25, 30)),
        ...     ]
        ... )
        >>> list(union_pipeline_1.run())
        [0, 1, 2, 3, 4, 25, 26, 27, 28, 29]

        You can use the ParallelPipeline to parallelize a group of pipes – each of which gets
        the same input.
        >>> pipeline_with_embedded_union = Pipeline([
        ...      ("pool", range(22)),
        ...      ("filters",  PipelineUnion([
        ...          ("div_5_filter", lambda x: filter(lambda i: i % 5 == 0, x)),
        ...          ("div_7_filter", lambda x: filter(lambda i: i % 7 == 0, x))
        ...         ]))
        ... ])
        >>> list(pipeline_with_embedded_union.run())
        [0, 5, 10, 15, 20, 0, 7, 14, 21]

    """

    def __call__(
        self,
        ex: Optional[_ExperimentalSequence] = None,
        **params,
    ) -> _ExperimentalSequence:
        """Pass the input values in parallel through the steps."""

        # Initialize the parameters objects.
        merged_params = self._merge_params_with_self_params(params)

        results = []

        # Run the parallel steps over the input
        for name, pipe in self.steps:
            all_params_for_step = merged_params.get(name, dict())
            if ex is None:
                if isinstance(pipe, Pool):
                    results.append(pipe(**all_params_for_step))
                elif isinstance(pipe, Iterable):
                    results.append(pipe)
                else:
                    raise NotImplementedError(
                        f"{pipe=} cannot be used in the PipelineUnion"
                    )
            else:
                assert isinstance(
                    pipe, Pipe
                ), f"{pipe=} is incompatible with the Pipe interface"
                results.append(pipe(ex, **all_params_for_step))

        union_results = chain.from_iterable(results)

        return union_results

    run = __call__


def _parse_params_to_nested_dict(params_dict: Dict, divider: str):
    """
    Converts a dictionary with a single level to a multi-level nested dictionary.

    Examples:
        >>> _parse_params_to_nested_dict({"a": 1}, divider="__")
        {'a': 1}
        >>> _parse_params_to_nested_dict({"a__b": 1, "a__c": 2}, divider="__")
        {'a': {'b': 1, 'c': 2}}
        >>> _parse_params_to_nested_dict(
        ...     {"a__b__alpha": 1, "a__b__beta": 2, "a__c__gamma": 3},
        ...     divider="__")
        {'a': {'b': {'alpha': 1, 'beta': 2}, 'c': {'gamma': 3}}}

        >>> _parse_params_to_nested_dict(
        ...     {"a:b:alpha": 1, "a:b:beta": 2, "a:c:gamma": 3},
        ...     divider=":")
        {'a': {'b': {'alpha': 1, 'beta': 2}, 'c': {'gamma': 3}}}
    """
    nested_dictionary: dict = copy.copy(params_dict)
    for key in params_dict.keys():
        if divider in key:
            value = nested_dictionary.pop(key)
            new_key, new_subkey = key.split(divider, 1)
            subdictionary = nested_dictionary.get(new_key, {})
            subdictionary.update({new_subkey: value})
            nested_dictionary[new_key] = subdictionary

    for key, value in nested_dictionary.items():
        if isinstance(value, dict):
            nested_dictionary[key] = _parse_params_to_nested_dict(
                value, divider=divider
            )

    return nested_dictionary


def make_pipeline(
    steps: Optional[Sequence[Union[Pool, Pipe]]] = None,
    params: Optional[Dict[str, Any]] = None,
    kind: Literal["serial", "union"] = "serial",
) -> Pipeline:
    """
    A factory function to make pipeline objects.

    The pipe objects' names will be set to the lowercase of their types, plus an index
    starting from 0 for non-unique names.

    Args:
        steps: a sequence of Pipe-compatible objects
        params: a dictionary of parameters passed to each Pipe by its inferred name
        kind: whether the steps should run in "serial", passing data from one to the next,
            or in "union", where all the steps get the same data and the output is the union
            of all the results.

    Returns:
        A pipeline object

    Examples:

        You can create pipelines using purely anonymous functions:
        >>> from itertools import product
        >>> make_pipeline([lambda: product(range(5), ["a", "b"])]) # doctest: +ELLIPSIS
        Pipeline(steps=[('<lambda>', <function <lambda> at 0x...>)], params={})

        You can create pipelines with normal functions.
        >>> def ab_pool(maximum=5): return product(range(maximum), ["a", "b"])
        >>> def even_filter(values): return filter(lambda i: i[0] % 2 == 0, values)
        >>> make_pipeline([ab_pool, even_filter]) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Pipeline(steps=[('ab_pool', <function ab_pool at 0x...>), \
        ('even_filter', <function even_filter at 0x...>)], params={})

        You can create pipelines with generators as their first elements functions.
        >>> ab_pool_gen = product(range(3), ["a", "b"])
        >>> pl = make_pipeline([ab_pool_gen, even_filter])
        >>> pl # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Pipeline(steps=[('step', <itertools.product object at 0x...>),
        ('even_filter', <function even_filter at 0x...>)], params={})
        >>> list(pl.run())
        [(0, 'a'), (0, 'b'), (2, 'a'), (2, 'b')]

        You can pass parameters into the different steps of the pl using the "params"
        argument:
        >>> def divisor_filter(x, divisor): return filter(lambda i: i[0] % divisor == 0, x)
        >>> pl = make_pipeline([ab_pool, divisor_filter],
        ... params = {"ab_pool": {"maximum":5}, "divisor_filter": {"divisor": 2}})
        >>> pl # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Pipeline(steps=[('ab_pool', <function ab_pool at 0x...>), \
        ('divisor_filter', <function divisor_filter at 0x...>)], \
        params={'ab_pool': {'maximum': 5}, 'divisor_filter': {'divisor': 2}})

        You can evaluate the pipeline means calling its `run` method:
        >>> list(pl.run())
        [(0, 'a'), (0, 'b'), (2, 'a'), (2, 'b'), (4, 'a'), (4, 'b')]

        ... or calling it directly:
        >>> list(pl())
        [(0, 'a'), (0, 'b'), (2, 'a'), (2, 'b'), (4, 'a'), (4, 'b')]

        You can update the parameters and evaluate again, giving different results:
        >>> pl.params = {"ab_pool": {"maximum": 7}, "divisor_filter": {"divisor": 3}}
        >>> list(pl())
        [(0, 'a'), (0, 'b'), (3, 'a'), (3, 'b'), (6, 'a'), (6, 'b')]

        If the pipeline needs parameters, then removing them will break the pipeline:
        >>> pl.params = {}
        >>> list(pl()) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: divisor_filter() missing 1 required positional argument: 'divisor'

        If multiple steps have the same inferred name, then they are given a suffix automatically,
        which has to be reflected in the params if used:
        >>> pl = make_pipeline([ab_pool, divisor_filter, divisor_filter])
        >>> pl.params = {
        ...     "ab_pool": {"maximum": 22},
        ...     "divisor_filter_0": {"divisor": 3},
        ...     "divisor_filter_1": {"divisor": 7}
        ... }
        >>> list(pl())
        [(0, 'a'), (0, 'b'), (21, 'a'), (21, 'b')]

        You can also use "partial" functions to include Pipes with defaults in the pipeline.
        Because the `partial` function doesn't inherit the __name__ of the original function,
        these steps are renamed to "step".
        >>> from functools import partial
        >>> pl = make_pipeline([partial(ab_pool, maximum=100)])
        >>> pl # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Pipeline(steps=[('step', functools.partial(<function ab_pool at 0x...>, maximum=100))], \
        params={})

        If there are multiple steps with the same name, they get suffixes as usual:
        >>> pl = make_pipeline([partial(range, stop=10), partial(divisor_filter, divisor=3)])
        >>> pl # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Pipeline(steps=[('step_0', functools.partial(<class 'range'>, stop=10)), \
        ('step_1', functools.partial(<function divisor_filter at 0x...>, divisor=3))], \
        params={})

        It is possible to create parallel pipelines too:
        >>> pl = make_pipeline([range(5), range(10,15)], kind="union")
        >>> pl
        PipelineUnion(steps=[('step_0', range(0, 5)), ('step_1', range(10, 15))], params={})

        >>> list(pl.run())
        [0, 1, 2, 3, 4, 10, 11, 12, 13, 14]

    """

    if steps is None:
        steps = []
    steps_: List[_StepType] = []
    raw_names_ = [getattr(pipe, "__name__", "step").lower() for pipe in steps]
    names_tally_ = dict([(name, raw_names_.count(name)) for name in set(raw_names_)])
    names_index_ = dict([(name, 0) for name in set(raw_names_)])

    for name, pipe in zip(raw_names_, steps):
        assert isinstance(pipe, get_args(Union[Pipe, Pool, Iterable]))

        if names_tally_[name] > 1:
            current_index_for_this_name = names_index_.get(name, 0)
            name_in_pipeline = f"{name}_{current_index_for_this_name}"
            names_index_[name] += 1
        else:
            name_in_pipeline = name

        steps_.append((name_in_pipeline, pipe))

    if kind == "serial":
        pipeline = Pipeline(steps_, params=params)
    elif kind == "union":
        pipeline = PipelineUnion(steps_, params=params)
    else:
        raise NotImplementedError(f"{kind=} is not implemented")

    return pipeline


class _ExperimentalCondition:
    """An _ExperimentalCondition represents a trial."""

    pass


_ExperimentalSequence = Iterable[_ExperimentalCondition]
_ExperimentalSequence.__doc__ = """
An _ExperimentalSequence represents a series of trials.
"""



import importlib
import logging
import pathlib
from typing import Optional

import typer
from typing_extensions import Annotated

from autora.serializer import (
    SupportedSerializer,
    default_serializer,
    dump_state,
    load_state,
)

_logger = logging.getLogger(__name__)


def main(
    fully_qualified_function_name: Annotated[
        str,
        typer.Argument(
            help="Fully qualified name of the function to load, like `module.function`"
        ),
    ],
    in_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(help="Path to a file with the initial state"),
    ] = None,
    in_loader: Annotated[
        SupportedSerializer,
        typer.Option(
            help="(de)serializer to load the data",
        ),
    ] = default_serializer,
    out_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(help="Path to output the final state"),
    ] = None,
    out_dumper: Annotated[
        SupportedSerializer,
        typer.Option(
            help="serializer to save the data",
        ),
    ] = default_serializer,
    verbose: Annotated[bool, typer.Option(help="Turns on info logging level.")] = False,
    debug: Annotated[bool, typer.Option(help="Turns on debug logging level.")] = False,
):
    """Run an arbitrary function on an optional input State object and save the output."""
    _configure_logger(debug, verbose)
    starting_state = load_state(in_path, in_loader)
    _logger.info(f"Starting State: {starting_state}")
    function = _load_function(fully_qualified_function_name)
    ending_state = function(starting_state)
    _logger.info(f"Ending State: {ending_state}")
    dump_state(ending_state, out_path, out_dumper)

    return


def _configure_logger(debug, verbose):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        _logger.debug("using DEBUG logging level")
    if verbose:
        logging.basicConfig(level=logging.INFO)
        _logger.info("using INFO logging level")


def _load_function(fully_qualified_function_name: str):
    """Load a function by its fully qualified name, `module.function_name`"""
    _logger.debug(f"_load_function: Loading function {fully_qualified_function_name}")
    module_name, function_name = fully_qualified_function_name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    function = getattr(module, function_name)
    _logger.debug(f"_load_function: Loaded function {function} from {module}")
    return function


if __name__ == "__main__":
    typer.run(main)



import pandas as pd
from sklearn.linear_model import LinearRegression

from autora.experimentalist.grid import grid_pool
from autora.state import StandardState, estimator_on_state, on_state
from autora.variable import Variable, VariableCollection


def initial_state(_):
    state = StandardState(
        variables=VariableCollection(
            independent_variables=[Variable(name="x", allowed_values=range(100))],
            dependent_variables=[Variable(name="y")],
            covariates=[],
        ),
        conditions=None,
        experiment_data=pd.DataFrame({"x": [], "y": []}),
        models=[],
    )
    return state


experimentalist = on_state(grid_pool, output=["conditions"])

experiment_runner = on_state(
    lambda conditions: conditions.assign(y=2 * conditions["x"] + 0.5),
    output=["experiment_data"],
)

theorist = estimator_on_state(LinearRegression(fit_intercept=True))



import logging
import pathlib
import tempfile
import uuid
from functools import partial

import hypothesis.strategies as st

from autora.serializer import SupportedSerializer, load_serializer

logger = logging.getLogger(__name__)


def _load_dump_via_string(o, module_name):
    s = load_serializer(module_name)
    return s.module.loads(s.module.dumps(o))


def _load_dump_via_disk(o, module_name):
    s = load_serializer(module_name)
    with tempfile.TemporaryDirectory() as tempdir:
        filename = str(uuid.uuid1())
        with open(pathlib.Path(tempdir, filename), f"w{s.file_mode}") as f:
            s.module.dump(o, f)
        with open(pathlib.Path(tempdir, filename), f"r{s.file_mode}") as f:
            o_loaded = s.module.load(f)
    return o_loaded


# Preload all the serializer modules to avoid delays when running each the first time (which could
# cause issues for hypothesis)
[load_serializer(s) for s in SupportedSerializer]

# Explicitly list all the strategies we use
load_dump_pickle_string = partial(_load_dump_via_string, module_name="pickle")
load_dump_pickle_disk = partial(_load_dump_via_disk, module_name="pickle")
load_dump_dill_string = partial(_load_dump_via_string, module_name="dill")
load_dump_dill_disk = partial(_load_dump_via_disk, module_name="dill")
load_dump_yaml_string = partial(_load_dump_via_string, module_name="yaml")
load_dump_yaml_disk = partial(_load_dump_via_disk, module_name="yaml")


serializer_dump_load_strategy = st.sampled_from(
    [
        load_dump_pickle_string,
        load_dump_dill_string,
        load_dump_yaml_string,
        load_dump_pickle_disk,
        load_dump_dill_disk,
        load_dump_yaml_disk,
    ]
)

if __name__ == "__main__":
    o = list("abcde")
    loader_dumper_disk = serializer_dump_load_strategy.example()
    o_loaded = loader_dumper_disk(o)
    print(o, o_loaded)



import logging
from typing import Optional, Sequence

import numpy as np
import pandas as pd
import sklearn.dummy
import sklearn.linear_model
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from hypothesis.extra import numpy as st_np
from hypothesis.extra import pandas as st_pd

from autora.state import StandardState
from autora.variable import ValueType, Variable, VariableCollection

VALUE_TYPE_DTYPE_MAPPING = {
    ValueType.BOOLEAN: bool,
    ValueType.INTEGER: int,
    ValueType.REAL: float,
    ValueType.SIGMOID: float,
    ValueType.PROBABILITY: float,
    ValueType.PROBABILITY_SAMPLE: float,
    ValueType.PROBABILITY_DISTRIBUTION: float,
    ValueType.CLASS: str,
}

logger = logging.getLogger(__name__)

AVAILABLE_SKLEARN_MODELS_STRATEGY = st.sampled_from(
    [
        sklearn.dummy.DummyRegressor,
        sklearn.linear_model.LinearRegression,
        sklearn.linear_model.Ridge,
        sklearn.linear_model.BayesianRidge,
    ]
)


@st.composite
def _name_label_units_strategy(draw, name=None, label=None, units=None, covariate=None):
    if name is None:
        name = draw(st.text(min_size=1))
    if label is None:
        label = draw(st.text(min_size=0))
    if units is None:
        units = draw(st.text(min_size=0))
    if covariate is None:
        covariate = draw(st.booleans())
    return name, label, units, covariate


@st.composite
def variable_boolean_strategy(draw, name=None, label=None, units=None, covariate=None):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.BOOLEAN
    allowed_values = [True, False]
    value_range = None
    rescale = 1
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


@st.composite
def variable_integer_strategy(draw, name=None, label=None, units=None, covariate=None):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.INTEGER

    value_range = draw(
        st.one_of(
            st.none(),
            st.tuples(st.integers(), st.integers())
            .filter(lambda x: x[0] != x[1])
            .map(sorted),
        )
    )
    if value_range is None:
        allowed_values = draw(st.one_of(st.none(), st.sets(st.integers(), min_size=1)))
    else:
        allowed_values = None

    rescale = draw(
        st.one_of(
            st.just(1),
            st.integers(),
            st.floats(allow_infinity=False, allow_subnormal=False, allow_nan=False),
        )
    )
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


@st.composite
def variable_real_strategy(draw, name=None, label=None, units=None, covariate=None):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.REAL
    range_strategy = st.floats(allow_nan=False, allow_subnormal=False)
    value_range = draw(
        st.one_of(
            st.none(),
            st.tuples(range_strategy, range_strategy)
            .filter(lambda x: x[0] != x[1])
            .map(sorted),
        )
    )

    if value_range is None:
        allowed_values = draw(st.one_of(st.none(), st.sets(range_strategy, min_size=1)))
    else:
        allowed_values = None
    rescale = draw(st.one_of(st.just(1), range_strategy))
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


@st.composite
def variable_probability_strategy(
    draw, name=None, label=None, units=None, covariate=None
):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.PROBABILITY
    value_range = (0, 1)
    allowed_values = None
    rescale = 1
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


@st.composite
def variable_probability_sample_strategy(
    draw, name=None, label=None, units=None, covariate=None
):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.PROBABILITY_SAMPLE
    value_range = (0, 1)
    allowed_values = None
    rescale = 1
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


@st.composite
def variable_probability_distribution_strategy(
    draw, name=None, label=None, units=None, covariate=None
):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.PROBABILITY_DISTRIBUTION
    value_range = (0, 1)
    allowed_values = None
    rescale = 1
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


@st.composite
def variable_sigmoid_strategy(draw, name=None, label=None, units=None, covariate=None):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.SIGMOID
    value_range = (-np.inf, +np.inf)
    allowed_values = None
    rescale = 1
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


@st.composite
def variable_class_strategy(draw, name=None, label=None, units=None, covariate=None):
    name, label, units, covariate = draw(
        _name_label_units_strategy(
            name=name, label=label, units=units, covariate=covariate
        )
    )
    value_type = ValueType.CLASS
    value_range = None
    rescale = 1
    allowed_values = draw(st.lists(st.text(min_size=1, max_size=16), unique=True))
    return Variable(
        name=name,
        variable_label=label,
        units=units,
        type=value_type,
        is_covariate=covariate,
        value_range=value_range,
        allowed_values=allowed_values,
        rescale=rescale,
    )


VARIABLE_STRATEGIES = (
    variable_boolean_strategy,
    variable_integer_strategy,
    variable_probability_strategy,
    variable_probability_sample_strategy,
    variable_probability_distribution_strategy,
    variable_sigmoid_strategy,
    variable_real_strategy,
    variable_class_strategy,
)


@st.composite
def variable_strategy(
    draw, elements=VARIABLE_STRATEGIES, value_type: Optional[ValueType] = None, **kwargs
):
    if value_type is None:
        strategy = draw(st.sampled_from(elements))

    else:
        strategy = {
            ValueType.BOOLEAN: variable_boolean_strategy,
            ValueType.INTEGER: variable_integer_strategy,
            ValueType.REAL: variable_real_strategy,
            ValueType.SIGMOID: variable_sigmoid_strategy,
            ValueType.PROBABILITY: variable_probability_strategy,
            ValueType.PROBABILITY_SAMPLE: variable_probability_sample_strategy,
            ValueType.PROBABILITY_DISTRIBUTION: variable_probability_distribution_strategy,
            ValueType.CLASS: variable_class_strategy,
        }[value_type]
    return draw(strategy(**kwargs))


@given(variable_strategy())
def test_variable_strategy_creation(o):
    assert o


@st.composite
def variablecollection_strategy(
    draw,
    elements=VARIABLE_STRATEGIES,
    value_type: Optional[ValueType] = None,
    max_ivs=5,
    max_dvs=1,
    max_covariates=2,
    name_max_length=32,
    **kwargs,
):
    n_ivs, n_dvs, n_covariates = draw(
        st.tuples(
            st.integers(min_value=1, max_value=max_ivs),
            st.integers(min_value=1, max_value=max_dvs),
            st.integers(min_value=0, max_value=max_covariates),
        )
    )

    n_variables = n_ivs + n_dvs + n_covariates

    names = draw(
        st.lists(
            st.text(min_size=1, max_size=name_max_length),
            unique=True,
            min_size=n_variables,
            max_size=n_variables,
        )
    )
    independent_variables = [
        draw(
            variable_strategy(
                name=names.pop(), value_type=value_type, elements=elements, **kwargs
            )
        )
        for _ in range(n_ivs)
    ]
    dependent_variables = [
        draw(
            variable_strategy(
                name=names.pop(), value_type=value_type, elements=elements, **kwargs
            )
        )
        for _ in range(n_dvs)
    ]
    covariates = [
        draw(
            variable_strategy(
                name=names.pop(), value_type=value_type, elements=elements, **kwargs
            )
        )
        for _ in range(n_covariates)
    ]

    vc = VariableCollection(
        independent_variables=independent_variables,
        dependent_variables=dependent_variables,
        covariates=covariates,
    )
    return vc


@given(variablecollection_strategy())
def test_variablecollection_strategy_creation(o):
    assert o


@st.composite
def dataframe_strategy(
    draw,
    variables: Optional[Sequence[Variable]] = None,
    value_type: Optional[ValueType] = None,
):
    if variables is None:
        variable_collection = draw(variablecollection_strategy(value_type=value_type))
        variables = (
            variable_collection.independent_variables
            + variable_collection.dependent_variables
            + variable_collection.covariates
        )

    df: pd.DataFrame = draw(
        st_pd.data_frames(
            columns=[
                st_pd.column(name=v.name, dtype=VALUE_TYPE_DTYPE_MAPPING[v.type])
                for v in variables
            ],
        )
    )

    return df


@given(dataframe_strategy())
def test_dataframe_strategy_creation(o):
    assert o is not None


@st.composite
def model_strategy(draw, models=AVAILABLE_SKLEARN_MODELS_STRATEGY):
    model = draw(models)

    n_x = draw(st.integers(min_value=1, max_value=5))
    n_y = draw(st.integers(min_value=1, max_value=1))
    n_measurements = draw(st.integers(min_value=5, max_value=100))

    elements = st_np.from_dtype(
        np.dtype(float),
        allow_infinity=False,
        allow_subnormal=False,
        allow_nan=False,
        # Include some reasonable extreme values. Values near the upper limit of the float
        # ~10**308, and very small values broke the fitting
        min_value=-1e5,
        max_value=1e5,
        min_magnitude=1e-3,
    )
    X = draw(st_np.arrays(float, shape=(n_measurements, n_x), elements=elements))
    y = draw(st_np.arrays(float, shape=(n_measurements, n_y), elements=elements))

    result = model().fit(X, y.ravel())
    return result


@given(model_strategy())
def test_model_strategy_creation(o):
    assert o


@st.composite
def standard_state_strategy(draw):
    variable_collection: VariableCollection = draw(variablecollection_strategy())
    conditions = draw(
        dataframe_strategy(variables=variable_collection.independent_variables)
    )
    experiment_data = draw(
        dataframe_strategy(
            variables=(
                variable_collection.independent_variables
                + variable_collection.dependent_variables
                + variable_collection.covariates
            )
        )
    )
    models = draw(st.lists(model_strategy(), min_size=0, max_size=5))
    s = StandardState(
        variables=variable_collection,
        conditions=conditions,
        experiment_data=experiment_data,
        models=models,
    )
    return s


@settings(suppress_health_check={HealthCheck.too_slow})
@given(standard_state_strategy())
def test_standard_state_strategy_creation(o):
    assert o


if __name__ == "__main__":
    print(model_strategy().example())



"""A submodule which handles importing supported serializers."""

import importlib
import logging
import pathlib
from collections import namedtuple
from enum import Enum
from typing import Dict, Optional, Union

from autora.state import State

_logger = logging.getLogger(__name__)

# Developer notes:
# Add a new serializer by:
# - including its name in the SupportedSerializer Enum. This {name} will be available on the
#   command line: `python -m autora.workflow --serializer {name}`
# - Adding the basic data about it to the _SERIALIZER_INFO dictionary. This should include the
#   fully qualified name under which it can be imported, and which file mode it expects:
#     "" for regular [i.e. files will be opened as `open(filename, "w")` for writing]
#     "b" for binary [i.e. files will be opened as `open(filename, "wb")` for writing]

_SERIALIZER_INFO_ENTRY = namedtuple(
    "_SERIALIZER_INFO_ENTRY", ["fully_qualified_module_name", "file_mode"]
)
LOADED_SERIALIZER = namedtuple("LOADED_SERIALIZER", ["module", "file_mode"])


class SupportedSerializer(str, Enum):
    """Listing of allowed serializers."""

    pickle = "pickle"
    dill = "dill"
    yaml = "yaml"


# Dictionary of details about each serializer
_SERIALIZER_INFO: Dict[SupportedSerializer, _SERIALIZER_INFO_ENTRY] = {
    SupportedSerializer.pickle: _SERIALIZER_INFO_ENTRY("pickle", "b"),
    SupportedSerializer.dill: _SERIALIZER_INFO_ENTRY("dill", "b"),
    SupportedSerializer.yaml: _SERIALIZER_INFO_ENTRY("autora.serializer.yaml_", ""),
}

# Set the default serializer for the package
default_serializer = SupportedSerializer.pickle

# A dictionary to handle lazy loading of the serializers
_LOADED_SERIALIZERS: Dict[SupportedSerializer, LOADED_SERIALIZER] = dict()


def load_serializer(
    serializer: Union[SupportedSerializer, str] = default_serializer
) -> LOADED_SERIALIZER:
    """
    Load a serializer, returning an object which includes data on the file mode it expects
    (regular or binary).

    Examples:
        The default serializer is pickle:
        >>> load_serializer()  # doctest: +ELLIPSIS
        LOADED_SERIALIZER(module=<module 'pickle' from '...'>, file_mode='b')

        All supported serializers can be loaded explictly:
        >>> p_ = load_serializer("pickle")
        >>> p_  # doctest: +ELLIPSIS
        LOADED_SERIALIZER(module=<module 'pickle' from '...'>, file_mode='b')

        >>> d_ = load_serializer("dill")
        >>> d_  # doctest: +ELLIPSIS
        LOADED_SERIALIZER(module=<module 'dill' from '...'>, file_mode='b')

        >>> y_ = load_serializer("yaml")
        >>> y_  # doctest: +ELLIPSIS
        LOADED_SERIALIZER(module=<module 'autora.serializer.yaml_' from '...'>, file_mode='')

        Note that the yaml serializer is a wrapped version of the `pyyaml` package,
        which conforms to the same interface as `pickle`.

        This function loads the modules lazily and caches them, and returns the same object if the
        same serializer is loaded more than once.
        >>> p_ is load_serializer("pickle")
        True
        >>> d_ is load_serializer("dill")
        True
        >>> y_ is load_serializer("yaml")
        True

        The Serializer can be specified by the SerializersSupported enum:
        >>> load_serializer(SupportedSerializer.pickle)  # doctest: +ELLIPSIS
        LOADED_SERIALIZER(module=<module 'pickle'...

        >>> load_serializer(SupportedSerializer.dill)  # doctest: +ELLIPSIS
        LOADED_SERIALIZER(module=<module 'dill'...

        >>> load_serializer(SupportedSerializer.yaml)  # doctest: +ELLIPSIS
        LOADED_SERIALIZER(module=<module 'autora.serializer.yaml_'...

    """
    serializer = SupportedSerializer(serializer)
    try:
        serializer_def = _LOADED_SERIALIZERS[serializer]

    except KeyError:
        serializer_info = _SERIALIZER_INFO[serializer]
        module_ = importlib.import_module(serializer_info.fully_qualified_module_name)
        serializer_def = LOADED_SERIALIZER(module_, serializer_info.file_mode)
        _LOADED_SERIALIZERS[serializer] = serializer_def

    return serializer_def


def load_state(
    path: Optional[pathlib.Path],
    loader: SupportedSerializer = default_serializer,
) -> Union[State, None]:
    """Load a State object from a path."""
    serializer = load_serializer(loader)
    if path is not None:
        _logger.debug(f"load_state: loading from {path=}")
        with open(path, f"r{serializer.file_mode}") as f:
            state_ = serializer.module.load(f)
    else:
        _logger.debug(f"load_state: {path=} -> returning None")
        state_ = None
    return state_


def dump_state(
    state_: State,
    path: Optional[pathlib.Path],
    dumper: SupportedSerializer = default_serializer,
) -> None:
    """Write a State object to a path."""
    serializer = load_serializer(dumper)
    if path is not None:
        _logger.debug(f"dump_state: dumping to {path=}")
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, f"w{serializer.file_mode}") as f:
            serializer.module.dump(state_, f)
    else:
        _logger.debug(f"dump_state: {path=} so writing to stdout")
        print(serializer.module.dumps(state_))
    return



import logging
from functools import wraps
from typing import Callable

_logger = logging.getLogger(__name__)


def deprecate(
    f: Callable,
    message: str,
    callback: Callable = _logger.warning,
):
    """
    Wrapper to make function aliases which print a warning that a name is an alias.

    Args:
        f: the function to be aliased
        message: the message to be emitted when the deprecated code is used
        callback: a function to call to handle the warning message

    Examples:
        >>> def original():
        ...     return 1
        >>> deprecated = deprecate(original, "`original` is deprecated.")

        The original function is unaffected:
        >>> original()
        1

        The aliased function works the same way, but also emits a warning.
        >>> deprecated()  # doctest: +SKIP
        `original` is deprecated.
        1

        You can also set a custom callback instead of the default "warning":
        >>> a0 = deprecate(original, "`original` is deprecated.", callback=print)
        >>> a0()
        `original` is deprecated.
        1
    """

    @wraps(f)
    def wrapper(*args, **kwds):
        callback(message)
        return f(*args, **kwds)

    return wrapper


def deprecated_alias(
    f: Callable, alias_name: str, callback: Callable = _logger.warning
):
    """
    Wrapper to make function aliases which print a warning that a name is an alias.

    Args:
        f: the function to be aliased
        alias_name: the name under which the function is aliased,
            like `foo = deprecated_alias(bar, "foo")
        callback: a function to call to handle the warning message

    Examples:
        >>> def original():
        ...     return 1
        >>> alias = deprecated_alias(original, "alias")

        The original function is unaffected:
        >>> original()
        1

        The aliased function works the same way, but also emits a warning.
        >>> alias()  # doctest: +SKIP
        Use `original` instead. `alias` is deprecated.
        1

        You can also set a custom callback instead of the default "warning":
        >>> a0 = deprecated_alias(original, "a0", callback=print)
        >>> a0()
        Use `original` instead. `a0` is deprecated.
        1

        The callback is given a single argument, the warning string.
        You can replace it if you like:
        >>> a0 = deprecated_alias(original, "a0", callback=lambda _: print("alternative message"))
        >>> a0()
        alternative message
        1
    """
    message = "Use `%s` instead. `%s` is deprecated." % (f.__name__, alias_name)
    wrapped = deprecate(f=f, message=message, callback=callback)
    return wrapped



#!/usr/bin/env python
import pathlib

import numpy as np
import pandas as pd
import typer
from matplotlib import pyplot as plt
from sklearn.model_selection import GridSearchCV

from autora.state import StandardState
from autora.workflow.__main__ import load_state

from .lib import ground_truth, noise_std


def plot_results(state: StandardState):
    x = np.linspace(-10, 10, 100).reshape((-1, 1))
    plt.plot(x, ground_truth(x), label="ground_truth", c="orange")
    plt.fill_between(
        x.flatten(),
        ground_truth(x).flatten() + noise_std,
        ground_truth(x).flatten() - noise_std,
        alpha=0.3,
        color="orange",
    )

    assert isinstance(state.experiment_data, pd.DataFrame)
    xi, yi = state.experiment_data["x"], state.experiment_data["y"]
    plt.scatter(xi, yi, label="observations")

    assert isinstance(state.models[-1], GridSearchCV)
    plt.plot(x, state.models[-1].predict(x), label="model")

    plt.xlabel("x")
    plt.ylabel("y")

    plt.legend()
    plt.show()


def main(filename: pathlib.Path):
    state = load_state(filename)
    assert isinstance(state, StandardState)
    plot_results(state)


if __name__ == "__main__":
    typer.run(main)



"""Tools to make grids of experimental conditions."""
from itertools import product

import pandas as pd

from autora.variable import VariableCollection


def pool(variables: VariableCollection) -> pd.DataFrame:
    """Creates exhaustive pool of conditions given a definition of variables with allowed_values.

    Args:
        variables: a VariableCollection with `independent_variables` – a sequence of Variable
            objects, each of which has an attribute `allowed_values` containing a sequence of
            values.

    Returns: a Result / Delta object with the conditions as a pd.DataFrame in the `conditions` field

    Examples:
        >>> from autora.state import State
        >>> from autora.variable import VariableCollection, Variable
        >>> from dataclasses import dataclass, field
        >>> import pandas as pd
        >>> import numpy as np

        With one independent variable "x", and some allowed values, we get exactly those values
        back when running the experimentalist:
        >>> pool(VariableCollection(
        ...     independent_variables=[Variable(name="x", allowed_values=[1, 2, 3])]
        ... ))
           x
        0  1
        1  2
        2  3

        The allowed_values must be specified:
        >>> pool(VariableCollection(independent_variables=[Variable(name="x")]))
        Traceback (most recent call last):
        ...
        AssertionError: grid_pool only supports independent variables with discrete...

        With two independent variables, we get the cartesian product:
        >>> pool(
        ...     VariableCollection(independent_variables=[
        ...         Variable(name="x1", allowed_values=[1, 2]),
        ...         Variable(name="x2", allowed_values=[3, 4]),
        ... ]))
           x1  x2
        0   1   3
        1   1   4
        2   2   3
        3   2   4

        If any of the variables have unspecified allowed_values, we get an error:
        >>> pool(
        ...     VariableCollection(independent_variables=[
        ...         Variable(name="x1", allowed_values=[1, 2]),
        ...         Variable(name="x2"),
        ... ]))
        Traceback (most recent call last):
        ...
        AssertionError: grid_pool only supports independent variables with discrete...


        We can specify arrays of allowed values:
        >>> pool(
        ...     VariableCollection(independent_variables=[
        ...         Variable(name="x", allowed_values=np.linspace(-10, 10, 101)),
        ...         Variable(name="y", allowed_values=[3, 4]),
        ...         Variable(name="z", allowed_values=np.linspace(20, 30, 11)),
        ... ]))
                 x  y     z
        0    -10.0  3  20.0
        1    -10.0  3  21.0
        2    -10.0  3  22.0
        3    -10.0  3  23.0
        4    -10.0  3  24.0
        ...    ... ..   ...
        2217  10.0  4  26.0
        2218  10.0  4  27.0
        2219  10.0  4  28.0
        2220  10.0  4  29.0
        2221  10.0  4  30.0
        <BLANKLINE>
        [2222 rows x 3 columns]

    """
    ivs = variables.independent_variables
    # Get allowed values for each IV
    l_iv_values = []
    l_iv_names = []
    for iv in ivs:
        assert iv.allowed_values is not None, (
            f"grid_pool only supports independent variables with discrete allowed values, "
            f"but allowed_values is None on {iv=} "
        )
        l_iv_values.append(iv.allowed_values)
        l_iv_names.append(iv.name)

    # Return Cartesian product of all IV values
    pool = product(*l_iv_values)
    conditions = pd.DataFrame(pool, columns=l_iv_names)

    return conditions


grid_pool = pool
"""Alias for pool"""



import logging

from hypothesis import given, settings
from hypothesis import strategies as st

from .test_serializer import serializer_dump_load_strategy
from .test_strategies import variable_strategy, variablecollection_strategy

logger = logging.getLogger(__name__)


@given(
    st.one_of(
        variable_strategy(),
        variablecollection_strategy(),
    ),
    serializer_dump_load_strategy,
)
@settings(deadline=1000)
def test_variable_serialize_deserialize(o, dump_load):
    o_loaded = dump_load(o)
    assert o_loaded == o
